import pandas as pd
import numpy as np
from shapely.geometry import LineString
import geopy.distance
from sklearn.cluster import DBSCAN
import more_itertools as mit
import multiprocessing as mp
from numba import njit
import skfuzzy as fuzz
from pandas.api.types import CategoricalDtype
import datetime
from haversine import haversine, Unit
from dateutil.parser import isoparse
import folium
from folium import plugins
import uuid
import random
import geopandas as gpd
from shapely.geometry import Point, LineString, MultiPoint
import math
from math import radians, sin, cos, sqrt, atan2


class GPSDataProcessor:
    """
    GPSDataProcessor is a class designed for processing GPS data, including mode detection and leg unification.

    This class provides methods for detecting user activities and modes of transportation in GPS data, as well as
    unifying individual legs and activities into meaningful segments.

    Attributes:
        - speed_th (float): A threshold for detecting walks based on speed.
        - acceleration_th (float): A threshold for detecting walks based on acceleration.
        - minimal_walking_duration (float): Minimum duration required to classify a segment as a walk.
        - minimal_trip_duration (float): Minimum duration required for a trip.
        - use_multiprocessing (bool): Flag indicating whether to use multiple processes for parallel processing.

    Example Usage:
        .. code-block:: python

            # Initialize GPSDataProcessor with custom Args
            processor = GPSDataProcessor(speed_th=1.5, acceleration_th=0.5, minimal_walking_duration=300, minimal_trip_duration=600, use_multiprocessing=True)

            poi_waypoints = data_processor.guess_home_work(waypoints, cell_size=0.3)
            smoothed_df = data_processor.smooth(poi_waypoints, sigma=100)
            segmented_df = data_processor.segment(smoothed_df)
            mode_df = data_processor.mode_detection(segmented_df)
            legs = data_processor.get_legs(df = mode_df)
    """

    def __init__(
        self,
        radius=0.5,
        min_samples=5,
        time_gap=850,
        speed_th=2.78,
        acceleration_th=0.5,
        minimal_walking_duration=100,
        minimal_trip_duration=120,
        use_multiprocessing=False,
    ):
        self.radius = radius
        #
        self.min_samples = min_samples
        self.time_gap = time_gap
        self.use_multiprocessing = use_multiprocessing
        #
        self.speed_th = speed_th
        self.acceleration_th = acceleration_th
        self.minimal_walking_duration = minimal_walking_duration
        self.minimal_trip_duration = minimal_trip_duration

    ########HOME WORK DETECTION
    def assign_cell(self, latitude, longitude, cell_size=0.2):
        """
        Assign a cell number to a GPS point based on latitude and longitude.

        Args:
            - latitude (float): Latitude of the GPS point.
            - longitude (float): Longitude of the GPS point.
            - cell_size (float): Size of the grid cell.

        Returns:
            - int: Cell number for the GPS point.
        """
        km_east = 111.320 * np.cos(np.deg2rad(latitude)) * longitude
        km_north = 110.574 * latitude
        x_index = km_east // cell_size
        y_index = km_north // cell_size
        cell_number = (1 / 2) * (y_index + x_index) * (y_index + x_index + 1) + x_index
        return cell_number

    def process_gps_data(self, waypoints, cell_size=0.2):
        """
        Process GPS data to detect home and work locations for users.

        Args:
            - waypoints (pandas.DataFrame): DataFrame with GPS waypoints.
            - cell_size (float): Size of the grid cell.

        Returns:
            - dict: A dictionary containing user IDs as keys and their detected home and work locations as values.
        """
        home_work_locations = {}

        for user_id in waypoints["user_id"].unique():
            user_waypoints = waypoints[waypoints["user_id"] == user_id]

            if user_waypoints.shape[0] != 0:
                user_waypoints_ = user_waypoints[
                    ["tracked_at", "latitude", "longitude"]
                ]

                if user_waypoints_.shape[0] != 0:
                    user_waypoints_["cell_number"] = self.assign_cell(
                        user_waypoints_["latitude"].values,
                        user_waypoints_["longitude"].values,
                        cell_size,
                    )
                    user_waypoints_["cell_number"] = user_waypoints_[
                        "cell_number"
                    ].astype("int64")

                    # Filter waypoints for workdays between 7:00 and 19:00
                    user_workdates = user_waypoints_[
                        (user_waypoints_["tracked_at"].dt.hour >= 7)
                        & (user_waypoints_["tracked_at"].dt.hour <= 19)
                        & (user_waypoints_["tracked_at"].dt.weekday < 5)
                    ]

                    # Calculate the most visited cells during workdays and overall for this user
                    most_visited_cells_workdates = (
                        user_workdates[["cell_number", "latitude", "longitude"]]
                        .groupby("cell_number")
                        .agg(["count", "mean"])
                    )
                    most_visited_cells_workdates.columns = (
                        most_visited_cells_workdates.columns.droplevel()
                    )
                    most_visited_cells_workdates.columns = [
                        "count",
                        "mean_lat",
                        "count_lon",
                        "mean_lon",
                    ]
                    del most_visited_cells_workdates["count_lon"]
                    most_visited_cells_workdates = (
                        most_visited_cells_workdates.reset_index()
                    )
                    most_visited_cells_workdates = most_visited_cells_workdates.rename(
                        columns={"mean_lat": "latitude", "mean_lon": "longitude"}
                    ).sort_values(by="count", ascending=False)

                    home_user = (
                        most_visited_cells_workdates.iloc[0]["latitude"],
                        most_visited_cells_workdates.iloc[0]["longitude"],
                    )
                    first_workdates = (
                        most_visited_cells_workdates.iloc[0]["latitude"],
                        most_visited_cells_workdates.iloc[0]["longitude"],
                    )

                    if most_visited_cells_workdates.shape[0] > 1:
                        second_workdates = (
                            most_visited_cells_workdates.iloc[1]["latitude"],
                            most_visited_cells_workdates.iloc[1]["longitude"],
                        )
                    else:
                        second_workdates = None

                    work_user = (
                        first_workdates
                        if (first_workdates != home_user)
                        else second_workdates
                    )

                    home_work_locations[user_id] = {
                        "Home_location": home_user,
                        "Work_location": work_user,
                    }

        return home_work_locations

    def guess_home_work(self, waypoints, cell_size=0.2):
        """
        Guess and assign home and work locations to waypoints based on processed GPS data.

        Args:
            - waypoints (pandas.DataFrame): DataFrame with GPS waypoints.
            - cell_size (float): Size of the grid cell.

        Returns:
            - pandas.DataFrame: DataFrame with 'home_loc' and 'work_loc' columns added.
        """
        home_work_data = self.process_gps_data(waypoints, cell_size)
        waypoints_ = waypoints.copy()
        waypoints_["home_loc"] = waypoints_["user_id"].map(
            lambda x: Point(home_work_data.get(x, {}).get("Home_location"))
        )
        waypoints_["work_loc"] = waypoints_["user_id"].map(
            lambda x: Point(home_work_data.get(x, {}).get("Work_location"))
        )
        return waypoints_

    ########SMOOTHING SECTION

    @staticmethod
    def smooth(
        df, accuracy_threshold=100, sigma=10, smoothing_method="time"
    ) -> pd.DataFrame:
        """
        Clean and preprocess a raw GPS points DataFrame using smoothing techniques.

        Args:
            - df (pandas.DataFrame): DataFrame with GPS points to be preprocessed.
            - accuracy_threshold (int, optional): Accuracy threshold for filtering. Defaults to 100.
            - sigma (int, optional): Sigma for Gaussian smoothing, defines the size of the smoothing window. Defaults to 10.
            - smoothing_method (str, optional): Smoothing method ('time' or 'space'). Defaults to 'time'.

        Returns:
            - pandas.DataFrame: Cleaned and preprocessed DataFrame.
        """
        processor = GPSDataProcessor()
        return processor._prepare(df, accuracy_threshold, sigma, smoothing_method)

    def _prepare(
        self, df, accuracy_threshold=100, sigma=10, smoothing_method="time"
    ) -> pd.DataFrame:
        """
        Internal method to prepare the DataFrame based on accuracy and smoothing method.

        Args:
            - df (pandas.DataFrame): The raw GPS points DataFrame.
            - accuracy_threshold (int, optional): Accuracy threshold for filtering. Defaults to 100.
            - sigma (int, optional): Sigma for Gaussian smoothing, defines the size of the smoothing window. Defaults to 10.
            - smoothing_method (str, optional): Smoothing method ('time' or 'space'). Defaults to 'time'.

        Returns:
            - pandas.DataFrame: Cleaned and preprocessed DataFrame.
        """
        unique_user_ids = df["user_id"].unique()
        cleaned_and_smoothed_df = pd.DataFrame()

        for user_id in unique_user_ids:
            user_df = df[df["user_id"] == user_id]  # Subsample by user_id
            if "accuracy" in user_df.columns:
                user_df = user_df[user_df.accuracy < accuracy_threshold]

            if smoothing_method == "time":
                user_df = self._time_based_smoothing(user_df, sigma)
            elif smoothing_method == "space":
                user_df = self._space_based_smoothing(user_df, sigma)

            cleaned_and_smoothed_df = pd.concat([cleaned_and_smoothed_df, user_df])

        cleaned_and_smoothed_df.reset_index(drop=True, inplace=True)
        return cleaned_and_smoothed_df

    @staticmethod
    def _time_based_smoothing(df, sigma=10) -> pd.DataFrame:
        """
        Apply time-based Gaussian smoothing on a DataFrame of GPS waypoints.

        Args:
            - df (pandas.DataFrame): DataFrame with points to be smoothed.
            - sigma (float): Smoothing parameter.

        Returns:
            - pandas.DataFrame: The smoothed DataFrame.
        """
        if df.shape[0] == 0:
            return df  # Return the input DataFrame if it's empty

        size = df.shape[0]
        output = df.reset_index(drop=True)
        output = output.sort_values("tracked_at", ascending=True)

        # Check if the data size is less than the smoothing window size
        if size <= 6:
            return output  # Return the input DataFrame without smoothing

        output["latitude"] = pd.to_numeric(output["latitude"])
        output["longitude"] = pd.to_numeric(output["longitude"])

        timestamps = (
            pd.to_datetime(output["tracked_at"]).astype("int64") / 10**9
        ).astype("int64")
        timestamps = timestamps.values
        longitudes = output["longitude"].values
        latitudes = output["latitude"].values

        output_longitudes = []
        output_latitudes = []

        sigma_squared = sigma**2

        for i in range(size):
            start = i - 5 if (i - 5 > 0) else 0
            end = i + 6 if (i + 6 < size) else size

            timestamp_df = timestamps[start:end]
            window_longitudes = longitudes[start:end]
            window_latitudes = latitudes[start:end]

            center_timestamp = timestamps[i]

            weights = timestamp_df - center_timestamp
            weights = weights**2
            weights = weights / sigma_squared
            weights = np.array(list(map(lambda x: math.exp(-x), weights)))
            sum_weights = np.sum(weights)

            new_longitude = np.sum(weights * window_longitudes) / sum_weights
            new_latitude = np.sum(weights * window_latitudes) / sum_weights

            output_longitudes.append(new_longitude)
            output_latitudes.append(new_latitude)

        output_longitudes = pd.Series(output_longitudes)
        output_latitudes = pd.Series(output_latitudes)

        output.longitude = output_longitudes
        output.latitude = output_latitudes

        return output

    def _space_based_smoothing(self, df, sigma=10) -> pd.DataFrame:
        """
        Apply space-based Gaussian smoothing on a DataFrame of GPS waypoints.

        Args:
            - df (pandas.DataFrame): DataFrame with points to be smoothed.
            - sigma (float): Smoothing parameter.

        Returns:
            - pandas.DataFrame: The smoothed DataFrame.
        """

        def extract_coordinates(line):
            return list(line.coords)

        def gaussian_smoothing(coords):
            if len(coords) < 3:
                return coords

            smoothed_coords = []
            size = len(coords)

            for i in range(size):
                start = i - 1 if i > 0 else 0
                end = i + 2 if i < size - 2 else size

                window = coords[start:end]

                if len(window) == 3:
                    p1, p2, p3 = window
                    x1, y1 = p1
                    x2, y2 = p2
                    x3, y3 = p3

                    new_x = (x1 + 4 * x2 + x3) / 6
                    new_y = (y1 + 4 * y2 + y3) / 6

                    smoothed_coords.append((new_x, new_y))
                else:
                    smoothed_coords.append(coords[i])

            return LineString(smoothed_coords)

    ########SEGMENTING SECTION

    @staticmethod
    def haversine_distance(lat1, lon1, lat2, lon2):
        """
        Calculate the Haversine distance between two GPS coordinates.

        Args:
            - lat1 (float): Latitude of the first point.
            - lon1 (float): Longitude of the first point.
            - lat2 (float): Latitude of the second point.
            - lon2 (float): Longitude of the second point.

        Returns:
            - float: Haversine distance in meters.
        """
        # Convert latitude and longitude from degrees to radians
        lat1, lon1, lat2, lon2 = map(radians, [lat1, lon1, lat2, lon2])

        # Haversine formula
        dlon = lon2 - lon1
        dlat = lat2 - lat1

        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * atan2(sqrt(a), sqrt(1 - a))

        # Radius of the Earth in meters (approximately)
        radius_ = 6371000  # meters

        # Calculate the distance in meters
        distance = radius_ * c
        return distance

    def create_route(self, df) -> pd.DataFrame:
        """
        Calculate additional statistics such as distance, time delta, speed, and acceleration for GPS waypoints.

        Args:
            - df (pandas.DataFrame): The waypoints DataFrame to be processed.

        Returns:
            - pandas.DataFrame: Waypoints DataFrame with additional statistics.
        """
        if df.shape[0] == 0:
            return pd.DataFrame(
                columns=[
                    "tracked_at_start",
                    "latitude_start",
                    "longitude_start",
                    "tracked_at_end",
                    "latitude_end",
                    "longitude_end",
                    "distance_meters",
                    "time_delta",
                    "speed",
                    "acceleration",
                    "detection",
                ]
            )

        res = df.reset_index().merge(
            df.shift(-1).dropna().reset_index(),
            on="index",
            suffixes=("_start", "_end"),
        )
        res = res.drop(
            columns=[
                "index",
                "user_id_end",
                "accuracy_start",
                "accuracy_end",
                "type_end",
            ],
            errors="ignore",
        )
        res.rename(
            columns={"user_id_start": "user_id", "type_start": "type"}, inplace=True
        )

        res["distance_meters"] = res.apply(
            lambda row: GPSDataProcessor.haversine_distance(
                row["latitude_start"],
                row["longitude_start"],
                row["latitude_end"],
                row["longitude_end"],
            ),
            axis=1,
        )

        res["time_delta"] = res.tracked_at_end - res.tracked_at_start
        res["time_delta"] = res["time_delta"].apply(
            lambda x: x.total_seconds()
        )  # in seconds
        res["speed"] = res["distance_meters"] / res["time_delta"]
        # This will be necessary for mode detection
        res["acceleration"] = res["speed"] / res["time_delta"]
        return res

    def prepare_for_detection(self, df) -> pd.DataFrame:
        """
        Flag detections as trips by default.

        Args:
            - df (pandas.DataFrame): Waypoints DataFrame to be flagged.

        Returns:
            - pandas.DataFrame: DataFrame with 'detection' column set to 'trip'.
        """
        df["detection"] = "trip"
        return df

    def activities_density(self, args):
        """
        Detect activities by density.

        Args:
            - args (tuple): A tuple containing the waypoints DataFrame and a clusterer object.

        Returns:
            - pandas.DataFrame: DataFrame with updated 'detection' column.
        """
        df, clusterer = args
        clusters_start = clusterer.fit_predict(
            np.radians(df[["latitude_start", "longitude_start"]])
        )

        df = df.copy()  # Create a copy of the DataFrame

        df.loc[:, "cluster_start"] = clusters_start.copy()
        df["cluster_end"] = df.cluster_start.shift(-1).fillna(0).astype("int")

        # update the detection column: Whenever the clusters column are different, we put trip in detection. Otherwise, we put activity.
        df.loc[
            (df["detection"] == "trip")
            & (df["cluster_start"] == df["cluster_end"])
            & (df["cluster_start"] != -1),
            ["detection"],
        ] = "activity"

        return df

    def correct_clusters(self, df):
        """
        Correct detected clusters by merging them based on time and speed.

        Args:
            - df (pandas.DataFrame): DataFrame with 'detection', 'speed', and 'time_delta' columns.
        """
        activity_indexes = df.loc[df.detection == "activity"].index.values
        trip_indexes = df.loc[df.detection == "trip"].index.values

        time_delta = df["time_delta"].values
        speed = df["speed"].values

        detection = df["detection"].values

        for group in mit.consecutive_groups(activity_indexes):
            indexes = list(group)
            window_time = np.sum(time_delta[indexes])
            if window_time < 120:
                detection[indexes] = "trip"

        for group in mit.consecutive_groups(trip_indexes):
            indexes = list(group)
            window_time = np.sum(time_delta[indexes])
            mean_speed = np.mean(speed[indexes])
            if window_time < 120 or mean_speed < 1:
                detection[indexes] = "activity"

        df["detection"] = detection

    def segment_per_user(self, df, user_id) -> pd.DataFrame:
        """
        Find clusters of waypoints for a specific user.

        Args:
            - df (pandas.DataFrame): Waypoints DataFrame to be processed.
            - user_id (int or str): The user_id for which to perform segmentation.

        Returns:
            - pandas.DataFrame: DataFrame with the segment starts and ends for the specified user.
        """
        # Filter the DataFrame for the specific user_id
        user_df = df[df["user_id"] == user_id]

        route_clusters_detected = pd.DataFrame()  # Initialize an empty DataFrame
        route_user = self.create_route(user_df)
        df = self.prepare_for_detection(route_user)

        df["day"] = df.tracked_at_start.apply(lambda x: x.day)
        df["month"] = df.tracked_at_start.apply(lambda x: x.month)
        df["year"] = df.tracked_at_start.apply(lambda x: x.year)

        db = DBSCAN(
            eps=self.radius / 6371.0,
            min_samples=self.min_samples,
            algorithm="ball_tree",
            metric="haversine",
        )

        route_clusters_detected = pd.DataFrame(
            columns=list(df.columns) + list(["cluster_start", "cluster_end"])
        )

        arguments = list(df.groupby(["day", "month", "year"]).groups.items())
        arguments = list(map(lambda x: (df.iloc[list(x[1])], db), arguments))

        if self.use_multiprocessing:
            pool = mp.Pool(processes=(mp.cpu_count() - 1))
            results = pool.map_async(self.activities_density, arguments)
            pool.close()
            pool.join()

            for res in results.get(timeout=1):
                route_clusters_detected = pd.concat(
                    [route_clusters_detected, res], ignore_index=True
                )
        else:
            for argument_list in arguments:
                res = self.activities_density(argument_list)
                route_clusters_detected = pd.concat(
                    [route_clusters_detected, res], ignore_index=True
                )

        route_clusters_detected = route_clusters_detected.sort_values(
            by="tracked_at_start", ascending=True
        )
        route_clusters_detected = route_clusters_detected.drop(
            columns=["day", "month", "year"]
        )
        route_clusters_detected = route_clusters_detected.reset_index().drop(
            columns="index"
        )

        self.correct_clusters(route_clusters_detected)

        route_clusters_detected = (
            route_clusters_detected.drop(
                route_clusters_detected[
                    route_clusters_detected.time_delta > self.time_gap
                ].index
            )
            .reset_index()
            .drop(columns="index")
        )

        route_clusters_detected = route_clusters_detected.drop(
            columns=[
                "cluster_start",
                "cluster_end",
                "type_end",
                "home_loc_end",
                "work_loc_end",
            ],
            errors="ignore",
        )

        route_clusters_detected.rename(
            columns={
                "user_id_start": "user_id",
                "type_start": "type",
                "home_loc_start": "home_loc",
                "work_loc_start": "work_loc",
            },
            inplace=True,
            errors="ignore",
        )

        return route_clusters_detected

    def segment(self, df) -> pd.DataFrame:
        """
        Segment waypoints for all unique user_ids in the DataFrame.

        Args:
            - df (pandas.DataFrame): Waypoints DataFrame with user_ids.

        Returns:
            - pandas.DataFrame: DataFrame with the segment starts and ends for all users.
        """
        unique_user_ids = df["user_id"].unique()
        result_df = pd.DataFrame()

        for user_id in unique_user_ids:
            user_segmentation = self.segment_per_user(df, user_id)
            result_df = pd.concat([result_df, user_segmentation], ignore_index=True)

        return result_df

    ##########MODE DETECTION SECTION

    def _detect_walks(self, args):
        """
        Tags the DataFrame at given indexes with 'walk' in the 'detection' column for the points corresponding to walks.

        Args:
            - args (tuple): A tuple containing the DataFrame `df`, `walk_speed_th`, `walk_acceleration_th`, `minimal_walk_duration`, and `minimal_trip_duration`.

        Returns:
            - pandas.DataFrame: DataFrame with 'walk' tags in the 'detection' column.
        """
        (
            df,
            walk_speed_th,
            walk_acceleration_th,
            minimal_walk_duration,
            minimal_trip_duration,
        ) = args
        i = 0
        window = []
        window_time = 0
        window_speed_median = 0
        window_acceleration_median = 0
        while i < len(df):
            condition_on_current_elem = (
                df.iloc[i]["speed"] <= walk_speed_th
                and df.iloc[i]["acceleration"] <= walk_acceleration_th
            )
            condition_on_existing_window = (
                window_speed_median <= walk_speed_th
                and window_acceleration_median <= walk_acceleration_th
            )

            if not condition_on_current_elem and not condition_on_existing_window:
                if len(window) and window_time >= minimal_walk_duration:
                    # Update walk status
                    trip_before_time = df.iloc[0 : window[0]]["time_delta"].sum()
                    trip_after_time = df.iloc[window[-1] + 1 :]["time_delta"].sum()

                    elements_before = window[0]
                    elements_after = len(df) - window[-1] + 1

                    if elements_before == 0 and elements_after == 0:
                        df.iloc[window, df.columns.get_loc("detection")] = "walk"
                    elif (
                        elements_before == 0
                        and trip_after_time >= minimal_trip_duration
                    ):
                        df.iloc[window, df.columns.get_loc("detection")] = "walk"
                    elif (
                        elements_after == 0
                        and trip_before_time >= minimal_trip_duration
                    ):
                        df.iloc[window, df.columns.get_loc("detection")] = "walk"
                    elif (
                        trip_before_time >= minimal_trip_duration
                        and trip_after_time >= minimal_trip_duration
                    ):
                        df.iloc[window, df.columns.get_loc("detection")] = "walk"

                    i = window[-1] + 1
                    window = []
                    window_time = 0
                    window_speed_median = 0
                    window_acceleration_median = 0
                else:
                    i += 1
                    window = []
                    window_time = 0
                    window_speed_median = 0
                    window_acceleration_median = 0
            else:
                window.append(i)
                window_time += df.iloc[i]["time_delta"]
                window_speed_median = df.iloc[window].speed.median()
                window_acceleration_median = df.iloc[window].acceleration.median()
                i += 1
        return df

    def _detect_modes(self, df):
        """
        Detects all modes except for walks using the fuzzy engine.

        Args:
            - df (pandas.DataFrame): DataFrame to be processed.

        Returns:
            - pandas.DataFrame: The modified DataFrame with estimated modes in the 'estimated_mode' column.
        """

        med_speed_verylow = [0, 0, 1.5, 2]
        med_speed_low = [1.5, 2, 4, 6]
        med_speed_medium = [5, 7, 11, 15]
        med_speed_high = [12, 15, 1000, 1000]

        acc95_low = [0, 0, 0.5, 0.6]
        acc95_medium = [0.5, 0.7, 1, 1.2]
        acc95_high = [1, 1.5, 1000, 1000]

        speed95_low = [0, 0, 6, 8]
        speed95_medium = [7.5, 9.5, 15, 18]
        speed95_high = [15, 20, 1000, 1000]

        possible_modes = []
        median_speed = np.median(df["speed"].values)
        acc_95per, speed_95per = (
            np.percentile(df["acceleration"].values, 95),
            np.percentile(df["speed"].values, 95),
        )

        medspeed_verylow_bool = (
            fuzz.trapmf(np.array([median_speed]), med_speed_verylow)[0] > 0
        )
        medspeed_low_bool = fuzz.trapmf(np.array([median_speed]), med_speed_low)[0] > 0
        medspeed_medium_bool = (
            fuzz.trapmf(np.array([median_speed]), med_speed_medium)[0] > 0
        )
        medspeed_high_bool = (
            fuzz.trapmf(np.array([median_speed]), med_speed_high)[0] > 0
        )

        acc95_low_bool = fuzz.trapmf(np.array([acc_95per]), acc95_low)[0] > 0
        acc95_medium_bool = fuzz.trapmf(np.array([acc_95per]), acc95_medium)[0] > 0
        acc95_high_bool = fuzz.trapmf(np.array([acc_95per]), acc95_high)[0] > 0

        speed95_low_bool = fuzz.trapmf(np.array([speed_95per]), speed95_low)[0] > 0
        speed95_medium_bool = (
            fuzz.trapmf(np.array([speed_95per]), speed95_medium)[0] > 0
        )
        speed95_high_bool = fuzz.trapmf(np.array([speed_95per]), speed95_high)[0] > 0

        # The paper suggests to take the minimum between the membership values.
        # We need to treat all cases where we have a non empty intersection
        if medspeed_verylow_bool and medspeed_low_bool:
            if (
                fuzz.trapmf(np.array([median_speed]), med_speed_verylow)[0]
                <= fuzz.trapmf(np.array([median_speed]), med_speed_low)[0]
            ):
                medspeed_low_bool = False
            else:
                medspeed_verylow_bool = False

        if medspeed_low_bool and medspeed_medium_bool:
            if (
                fuzz.trapmf(np.array([median_speed]), med_speed_low)[0]
                <= fuzz.trapmf(np.array([median_speed]), med_speed_medium)[0]
            ):
                medspeed_medium_bool = False
            else:
                medspeed_low_bool = False

        if medspeed_medium_bool and medspeed_high_bool:
            if (
                fuzz.trapmf(np.array([median_speed]), med_speed_medium)[0]
                <= fuzz.trapmf(np.array([median_speed]), med_speed_high)[0]
            ):
                medspeed_high_bool = False
            else:
                medspeed_medium_bool = False

        if acc95_low_bool and acc95_medium_bool:
            if (
                fuzz.trapmf(np.array([acc_95per]), acc95_low)[0]
                <= fuzz.trapmf(np.array([acc_95per]), acc95_medium)[0]
            ):
                acc95_medium_bool = False
            else:
                acc95_low_bool = False

        if acc95_medium_bool and acc95_high_bool:
            if (
                fuzz.trapmf(np.array([acc_95per]), acc95_medium)[0]
                <= fuzz.trapmf(np.array([acc_95per]), acc95_high)[0]
            ):
                acc95_high_bool = False
            else:
                acc95_medium_bool = False

        if speed95_low_bool and speed95_medium_bool:
            if (
                fuzz.trapmf(np.array([speed_95per]), speed95_low)[0]
                <= fuzz.trapmf(np.array([speed_95per]), speed95_medium)[0]
            ):
                speed95_medium_bool = False
            else:
                speed95_low_bool = False

        if speed95_medium_bool and speed95_high_bool:
            if (
                fuzz.trapmf(np.array([speed_95per]), speed95_medium)[0]
                <= fuzz.trapmf(np.array([speed_95per]), speed95_high)[0]
            ):
                speed95_high_bool = False
            else:
                speed95_medium_bool = False

        if medspeed_verylow_bool and acc95_low_bool:
            possible_modes.append("Walk")
        if medspeed_verylow_bool and acc95_medium_bool:
            possible_modes.append("Cycle")
        if medspeed_verylow_bool and acc95_high_bool:
            possible_modes.append("Cycle")

        if medspeed_low_bool and acc95_low_bool and speed95_low_bool:
            possible_modes.append("Cycle")
        if medspeed_low_bool and acc95_low_bool and speed95_medium_bool:
            possible_modes.append("Urban")
        if medspeed_low_bool and acc95_low_bool and speed95_high_bool:
            possible_modes.append("Car")
        if medspeed_low_bool and acc95_medium_bool:
            possible_modes.append("Urban")
        if medspeed_low_bool and acc95_high_bool and speed95_low_bool:
            possible_modes.append("Urban")
        if medspeed_low_bool and acc95_high_bool and speed95_medium_bool:
            possible_modes.append("Car")
        if medspeed_low_bool and acc95_high_bool and speed95_high_bool:
            possible_modes.append("Car")

        if medspeed_medium_bool and acc95_low_bool:
            possible_modes.append("Urban")
        if medspeed_medium_bool and acc95_medium_bool:
            possible_modes.append("Car")
        if medspeed_medium_bool and acc95_high_bool:
            possible_modes.append("Car")

        if medspeed_high_bool and acc95_low_bool:
            possible_modes.append("Rail")
        if medspeed_high_bool and acc95_medium_bool:
            possible_modes.append("Car")
        if medspeed_high_bool and acc95_high_bool:
            possible_modes.append("Car")

        # df['estimated_mode'] = ",".join(possible_modes)
        df.loc[:, "estimated_mode"] = ",".join(possible_modes)
        return df

    def mode_detection(self, df):
        """
        Perform mode detection on a DataFrame, including walk detection and other modes using fuzzy logic.

        Args:
            - df (pandas.DataFrame): DataFrame with GPS data.

        Returns:
            - pandas.DataFrame: DataFrame with modes and walk tags in the 'estimated_mode' and 'detection' columns.
        """
        df["estimated_mode"] = np.nan
        user_trips = df[df.detection == "trip"].index.values

        arguments = [list(x) for x in mit.consecutive_groups(user_trips)]
        arguments = list(
            map(
                lambda x: (
                    df.iloc[x],
                    self.speed_th,
                    self.acceleration_th,
                    self.minimal_walking_duration,
                    self.minimal_trip_duration,
                ),
                arguments,
            )
        )

        if self.use_multiprocessing:
            pool = mp.Pool(processes=(mp.cpu_count() - 1))
            results = pool.map_async(self._detect_walks, arguments)
            pool.close()
            pool.join()

            for res in results.get(timeout=1):
                df.loc[res.index] = res
        else:
            for argument_list in arguments:
                res = self._detect_walks(argument_list)
                df.loc[res.index] = res

        user_trips = df[df.detection == "trip"].index.values

        arguments = [list(x) for x in mit.consecutive_groups(user_trips)]
        arguments = list(map(lambda x: (df.iloc[x]), arguments))

        if self.use_multiprocessing:
            pool = mp.Pool(processes=(mp.cpu_count() - 1))
            results = pool.map_async(self._detect_modes, arguments)
            pool.close()
            pool.join()

            for res in results.get(timeout=1):
                df.loc[res.index] = res
        else:
            for argument_list in arguments:
                res = self._detect_modes(argument_list)
                df.loc[res.index] = res

        df.loc[df.detection == "walk", "estimated_mode"] = "Walk"

        return df

    ##########LEG UNIFICATION SECTION

    def get_legs(self, df):
        """
        Extract legs and activities from a DataFrame and aggregate the data for each leg or activity.

        Args:
            - df (pandas.DataFrame): DataFrame with GPS data, including 'detection', 'user_id', and 'type' columns.

        Returns:
            - pandas.DataFrame: DataFrame with aggregated legs and activities.
        """

        # Create a new grouping key based on consecutive values in the 'detection' column, 'user_id', and 'type'
        group_key = (
            (df["detection"] != df["detection"].shift())
            | (df["user_id"] != df["user_id"].shift())
            | (df["type"] != df["type"].shift())
        ).cumsum()

        # Define custom aggregation functions for activity and trip
        def aggregate_activity(group):
            first_start = group["tracked_at_start"].iloc[0]
            last_end = group["tracked_at_end"].iloc[-1]

            # Calculate the centroid of the activity points
            lon = group["longitude_start"].mean()
            lat = group["latitude_start"].mean()

            # Calculate the duration of the activity
            duration = pd.to_datetime(last_end) - pd.to_datetime(first_start)

            return pd.Series(
                {
                    "user_id": group["user_id"].iloc[0],
                    "type": group["type"].iloc[0],
                    "tracked_at_start": first_start,
                    "tracked_at_end": last_end,
                    "duration": duration,
                    "detection": group["detection"].iloc[0],
                    "longitude_start": lon,
                    "latitude_start": lat,
                    "longitude_end": lon,
                    "latitude_end": lat,
                    "distance": group["distance_meters"].sum(),
                    "time_delta": group["time_delta"].sum(),
                    "speed": group["speed"].mean(),
                    "acceleration": group["acceleration"].mean(),
                    "estimated_mode": group["estimated_mode"].iloc[0],
                    "geometry": Point(lat, lon),
                }
            )

        def aggregate_trip(group):
            first_start = group["tracked_at_start"].iloc[0]
            last_end = group["tracked_at_end"].iloc[-1]

            # Check if there are at least two points in the group
            if len(group) < 2:
                # Handle the case with only one point
                lon = group["longitude_start"].iloc[0]
                lat = group["latitude_start"].iloc[0]
            else:
                # Create a LineString from the trip points
                trip_line = LineString(
                    zip(group["latitude_start"], group["longitude_start"])
                )
                lon = group["longitude_start"].iloc[0]
                lat = group["latitude_start"].iloc[0]

            # Calculate the duration of the trip
            duration = pd.to_datetime(last_end) - pd.to_datetime(first_start)

            return pd.Series(
                {
                    "user_id": group["user_id"].iloc[0],
                    "type": group["type"].iloc[0],
                    "tracked_at_start": first_start,
                    "tracked_at_end": last_end,
                    "duration": duration,
                    "detection": group["detection"].iloc[0],
                    "longitude_start": lon,
                    "latitude_start": lat,
                    "longitude_end": group["longitude_end"].iloc[-1],
                    "latitude_end": group["latitude_end"].iloc[-1],
                    "distance": group["distance_meters"].sum(),
                    "time_delta": group["time_delta"].sum(),
                    "speed": group["speed"].mean(),
                    "acceleration": group["acceleration"].mean(),
                    "estimated_mode": group["estimated_mode"].iloc[0],
                    "geometry": trip_line if len(group) >= 2 else Point(lat, lon),
                }
            )

        # Group by the new key and apply the custom aggregation functions
        self.result = df.groupby([group_key, "detection", "user_id", "type"]).apply(
            lambda x: aggregate_activity(x)
            if x["detection"].iloc[0] == "activity"
            else aggregate_trip(x)
        )

        # Reset the index and drop the grouping key
        self.result = self.result.reset_index(level=[0, 1, 2, 3], drop=True)

        # Drop the 'longitude_start' and 'latitude_start' columns
        self.result = self.result.drop(
            ["longitude_start", "latitude_start", "longitude_end", "latitude_end"],
            axis=1,
        )

        self.result.loc[self.result.detection == "activity", "type"] = "Staypoint"
        self.result.loc[self.result.detection == "trip", "type"] = "Leg"

        # Add the home and work locations if available in df
        if "home_loc" in df and "work_loc" in df:
            # Both 'home_loc' and 'work_loc' columns are present
            self.result = pd.merge(
                self.result,
                df[["home_loc", "work_loc", "user_id"]].drop_duplicates(),
                on="user_id",
                how="left",
            )
        elif "home_loc" in df:
            # Only 'home_loc' is present
            self.result = pd.merge(
                self.result,
                df[["home_loc", "user_id"]].drop_duplicates(),
                on="user_id",
                how="left",
            )
        elif "work_loc" in df:
            # Only 'work_loc' is present
            self.result = pd.merge(
                self.result,
                df[["work_loc", "user_id"]].drop_duplicates(),
                on="user_id",
                how="left",
            )
        else:
            # Neither 'home_loc' nor 'work_loc' are present in df
            pass  # You can choose to ignore or perform another action

        return self.result



# Data Manipulation and Analysis
import pandas as pd
import geopandas as gpd
import numpy as np
import pyproj

# Visualization
import plotly.graph_objects as go
import seaborn as sns

import matplotlib.pyplot as plt
import contextily as ctx
from pointpats.centrography import (
    mean_center,
    weighted_mean_center,
    euclidean_median,
    dtot,
)
from shapely.geometry import MultiPoint
from matplotlib.patches import Ellipse
from scipy.spatial import distance
from shapely.geometry import Polygon, Point
import folium
import networkx as nx
import pointpats


class GPStoActionspace:
    """
    A class for analyzing GPS-based user activity and generating action spaces.

    This class provides methods to process GPS data, compute action spaces for users,
    visualize action spaces using Matplotlib and Folium, calculate covariance matrices,
    and compute innovation rates based on user activity motifs.
    """

    def __init__(self) -> None:
        plt.rcParams["figure.figsize"] = [5, 5]
        plt.rcParams["figure.dpi"] = 120  # 200 e.g. is really fine, but slower
        sns.set_theme(style="white")
        sns.set_palette("Pastel2_r")
        cm = sns.light_palette("#fff4ac", as_cmap=True)

    @staticmethod
    def verify_columns(df: pd.DataFrame, expected_columns: list[str]) -> None:
        """
        Verify if expected columns are present in the DataFrame.

        Args:
            - df (pandas.DataFrame): DataFrame to check.
            - expected_columns (list[str]): List of column names expected in the DataFrame.

        Raises:
            - ValueError: If columns are missing in the DataFrame.
        """

        missing_columns = [col for col in expected_columns if col not in df.columns]
        if missing_columns:
            raise ValueError(
                f"Columns missing in DataFrame: {', '.join(missing_columns)}"
            )

    @staticmethod
    def modified_std_distance(pp, center) -> float:
        """
        Calculate standard distance of a point array like std_distance() does in PySAL with the mean center, but we can specify here a different center.

        Args:
            - pp: point pattern
            - center: array([ x, y])

        Returns:
            - float: standard distance
        """

        return np.sqrt(
            1
            / len(pp)
            * (
                np.sum((pp.geometry.x - center[0]) ** 2)
                + np.sum((pp.geometry.y - center[1]) ** 2)
            )
        )

    def compute_action_space(
        self, act: pd.DataFrame, aggregation_method: str
    ) -> pd.DataFrame:
        """
        Compute action space Args for each user.

        Args:
            - act (pandas.DataFrame): DataFrame with user activity data.
            - aggregation_method (str): Aggregation method ('user_id' or 'user_id_day').

        Returns:
            - pandas.DataFrame: DataFrame with action space Args for each user.
        """

        action_space = []
        cols = [
            "user_id",
            "lon",
            "lat",
            "cluster_size",
            "cluster_info",
            "location_id",
            "home_location_id",
            "num_trip",
            "main_home_location_id",
            "dist_from_home",
            "imputed_purpose",
        ]

        self.verify_columns(act, cols)

        # check if aggregation method is user_id or user_id_day

        if aggregation_method not in ["user_id", "user_id_day"]:
            raise ValueError(
                f"aggreg_method must be either user_id or user_id_day, not {aggregation_method}"
            )

        for user_id_ in act[aggregation_method].unique():
            act_ = act.loc[act[aggregation_method] == user_id_, cols].copy()
            # get the main home id at user_id scale
            main_home_id = act_.home_location_id.value_counts().idxmax()

            if main_home_id == "not_detected":
                continue
            # elif np.any(act_.dist_from_home > 3000000):
            #    continue
            elif act_.main_home_location_id.unique() != act_.home_location_id.unique():
                continue
            elif np.all(act_.num_trip == 0):
                continue
            else:
                act_.drop(columns=["num_trip", "dist_from_home"], inplace=True)

                # Assuming your DataFrame is called df
                act_ = gpd.GeoDataFrame(
                    act_,
                    geometry=gpd.points_from_xy(act_["lon"], act_["lat"]),
                    crs="EPSG:4326",
                )

                # Change the CRS to EPSG:2056
                act_ = act_.to_crs(
                    "EPSG:2056"
                )  # This should be a parameter to change in the method, but default set to this one

                # Now, you can access the transformed x and y coordinates
                act_["x"], act_["y"] = act_.geometry.x, act_.geometry.y

                # And drop duplicated locations (information of visit count is in cluster_size)
                act_ = act_.drop_duplicates(ignore_index=True)

                # Differentiate most frequent locations from all locations
                act_freq = act_.loc[
                    act_.cluster_info.isin(["Most visited", "Frequent visit"])
                ].reset_index(drop=True)

                if len(act_) == 0:
                    continue
                else:
                    # Find mean center of a point array
                    mc_ = mean_center(act_[["x", "y"]])
                    # Find weighted mean center of a marked point pattern.
                    wmc_ = weighted_mean_center(act_[["x", "y"]], act_.cluster_size)

                    # Calculate Args of standard deviational ellipse for a point pattern
                    if len(act_) > 2:
                        sx, sy, theta = pointpats.centrography.ellipse(act_[["x", "y"]])
                        theta_degree = np.degrees(
                            theta
                        )  # need degree of rotation to plot the ellipse
                        # Calculate the surface of the SD ellipse of the action space, in square meters
                        surface_ellipse = np.pi * sx * sy
                    else:
                        surface_ellipse = 0
                    # sx, sy, theta, theta_degree

                    # Calculate the ellipse motplotlib objet
                    e = Ellipse(
                        xy=wmc_,
                        width=sx * 2,
                        height=sy * 2,
                        angle=-theta_degree,
                        alpha=0.2,
                        facecolor="#c9cfbc",
                    )  # angle is rotation in degrees (anti-clockwise)

                    # Find the main home coordinates
                    find_home_geom = act_.loc[
                        act_.location_id == act_.home_location_id.unique()[0],
                        "geometry",
                    ]
                    if len(find_home_geom) == 1:
                        home_loc = np.array(
                            [find_home_geom.x, find_home_geom.y], dtype=float
                        )
                    elif len(find_home_geom) > 1:
                        home_loc = np.array(
                            [find_home_geom[0].x, find_home_geom[0].y], dtype=float
                        )

                    # Sum of Euclidean distances between event points and a selected point.
                    sum_dist_home = dtot(home_loc, act_[["x", "y"]].to_numpy())

                    # Compute location regularity
                    n_all = len(act_)
                    n_freq = len(act_freq)
                    regularity = n_freq / n_all

                    # Calculate standard distance between all nodes and home
                    std_dist_all = self.modified_std_distance(act_, home_loc)
                    # Calculate standard distance between freq nodes and home
                    if len(act_freq) > 0:
                        std_dist_freq = self.modified_std_distance(act_freq, home_loc)
                    else:
                        std_dist_freq = np.nan

                    # Calculate the proximity
                    proximity = std_dist_freq / std_dist_all

                    # Calculate the Euclidean median for a point pattern.
                    median_ = euclidean_median(act_[["x", "y"]].to_numpy())[0]

                    # Get area of convex hull for frequent (i.e. habitual) locations in square meters
                    # And Coefficient of intensity
                    if len(act_freq.location_id.unique()) > 2:
                        freq_hull_area = MultiPoint(act_freq.geometry).convex_hull.area
                        intensity = (
                            MultiPoint(act_freq.geometry).convex_hull.area
                            / MultiPoint(act_.geometry).convex_hull.area
                        )
                    elif len(act_freq.location_id.unique()) == 2:
                        loc1 = (
                            act_freq[["x", "y"]]
                            .loc[
                                act_freq.location_id == act_freq.location_id.unique()[0]
                            ]
                            .drop_duplicates()
                        )
                        loc2 = (
                            act_freq[["x", "y"]]
                            .loc[
                                act_freq.location_id == act_freq.location_id.unique()[1]
                            ]
                            .drop_duplicates()
                        )

                        coord1 = loc1[["x", "y"]].to_numpy().flatten()
                        coord2 = loc2[["x", "y"]].to_numpy().flatten()

                        freq_hull_area = distance.euclidean(coord1, coord2)
                        intensity = 0
                    else:
                        freq_hull_area = 0
                        intensity = 0

                    # Calculate euclidean distance between weighted mean center and main home location in meters
                    home_shift = distance.euclidean(wmc_, home_loc.flatten())

                    action_space_ = np.array(
                        [
                            user_id_,
                            main_home_id,
                            sum_dist_home,
                            proximity,
                            std_dist_all,
                            std_dist_freq,
                            median_,
                            intensity,
                            regularity,
                            n_all,
                            n_freq,
                            freq_hull_area,
                            home_shift,
                            e,
                            surface_ellipse,
                            wmc_[0],
                            wmc_[1],
                            sx,
                            sy,
                            theta_degree,
                        ]
                    )
                    action_space.append(action_space_)

        columns_ = [
            aggregation_method,
            "main_home_id",
            "sum_dist_home",
            "proximity",
            "std_dist_all",
            "std_dist_freq",
            "median_",
            "intensity",
            "regularity",
            "n_all",
            "n_freq",
            "freq_hull_area",
            "home_shift",
            "ellipse_2056",
            "surface_ellipse",
            "wmc_x",
            "wmc_y",
            "sx",
            "sy",
            "theta_degree",
        ]
        action_space = pd.DataFrame(action_space, columns=columns_)
        action_space.set_index(aggregation_method, inplace=True)

        # Parse to float
        col_2 = [
            "sum_dist_home",
            "proximity",
            "std_dist_all",
            "std_dist_freq",
            "median_",
            "intensity",
            "regularity",
            "n_all",
            "n_freq",
            "freq_hull_area",
            "home_shift",
            "surface_ellipse",
        ]
        action_space[col_2] = action_space[col_2].astype(float)

        return action_space

    def plot_ellipses(
        self, action_space: pd.DataFrame, aggregation_method: str
    ) -> folium.Map:
        """
        Plot ellipses on a Folium map based on action space data.

        Args:
            - action_space (pandas.DataFrame): DataFrame with action space data.
            - aggregation_method (str): Aggregation method ('user_id' or 'user_id_day').

        Returns:
            - folium.Map: Folium map with ellipses.
        """

        # Create a list to store ellipse geometries
        ellipse_geometries = []

        # Define the source and target coordinate reference systems (CRS)
        source_crs = "EPSG:2056"
        target_crs = "EPSG:4326"

        # Iterate through the DataFrame and add ellipses to the list
        for idx, row in action_space.reset_index().iterrows():
            # Extract ellipse information
            user_id = row[aggregation_method]
            ellipse = row["ellipse_2056"]

            # Convert ellipse to Polygon geometry
            ellipse_polygon = Polygon(ellipse.get_verts())
            ellipse_coords = [[x, y] for x, y in ellipse_polygon.exterior.coords]

            ellipse_polygon = Polygon(ellipse.get_verts())
            ellipse_geometries.append(ellipse_polygon)

        # Create a GeoDataFrame from the list of ellipse geometries
        ellipse_gdf = gpd.GeoDataFrame(
            geometry=ellipse_geometries, crs=source_crs
        ).to_crs(target_crs)

        # Calculate the mean centroid of all ellipses
        mean_lat = ellipse_gdf.geometry.apply(lambda geom: geom.centroid.x).mean()
        mean_lon = ellipse_gdf.geometry.apply(lambda geom: geom.centroid.y).mean()

        # Create a Folium map centered around the mean centroid
        mymap = folium.Map(
            location=[mean_lon, mean_lat], zoom_start=12, control_scale=True
        )

        # Iterate through the GeoDataFrame and add ellipses to the map
        for idx, row in ellipse_gdf.iterrows():
            # Extract ellipse information
            user_id = idx  # Assuming user_id is the index in the GeoDataFrame
            ellipse_polygon = row["geometry"]

            # Convert ellipse to coordinates for Folium
            ellipse_coords = list(ellipse_polygon.exterior.coords)
            ellipse_coords_t = [[y, x] for x, y in ellipse_coords]

            # Add the ellipse to the map
            folium.Polygon(
                locations=ellipse_coords_t,
                color="blue",
                fill=True,
                fill_color="blue",
                fill_opacity=0.02,
                popup=f"User ID: {user_id}",
            ).add_to(mymap)

        # Display the map
        return mymap  # .save("ellipse_map.html")

    def covariance_matrix(
        self,
        action_space: pd.DataFrame,
        title: str = "",
        annot: bool = False,
        cmap=sns.diverging_palette(360, 65, l=80, as_cmap=True),
    ):
        """
        Map the correlation, covariance, or p-values of a set of observed variables.

        Args:
            - action_space (pandas.DataFrame): DataFrame with action space data.
            - title (str): Title for the heatmap.
            - annot (bool): If True, return the values in the heatmap cells.
            - cmap: Color map for the heatmap.

        Returns:
        - Heatmap of correlation, covariance, or p-values of observed variables.
        """

        cols = [
            "sum_dist_home",
            "proximity",
            "std_dist_all",
            "std_dist_freq",
            "median_",
            "intensity",
            "regularity",
            "n_all",
            "n_freq",
            "freq_hull_area",
            "home_shift",
            "surface_ellipse",
        ]

        self.verify_columns(action_space, cols)

        X = action_space[cols]

        X_ = X.corr()
        vmin = None
        title2 = "Correlation matrix of X_"

        # Generate a mask for the upper triangle
        mask = np.triu(np.ones_like(X_, dtype=bool), 0)

        # Set the title
        if len(title) < 1:
            title = title2  # pre-set titles else specified titles

        # Set up the matplotlib figure
        f, ax = plt.subplots(ncols=1, figsize=(7, 7))
        plt.title(title)

        # Draw the heatmap with the mask and correct aspect ratio
        fig = sns.heatmap(
            X_,
            mask=mask,
            cmap=cmap,
            vmin=vmin,
            vmax=None,
            center=0,
            square=True,
            linewidths=0.1,
            cbar_kws={"shrink": 0.7},
            annot=annot,
            fmt=".2g",
        )
        return plt.show()

    def plot_action_space(
        self,
        act: pd.DataFrame,
        action_space: pd.DataFrame,
        user: str,
        how: str = "vignette",
        save: bool = False,
    ):
        """
        Plot the action space for a specific user.

        Args:
            - act (pandas.DataFrame): DataFrame with user activity data.
            - action_space (pandas.DataFrame): DataFrame with action space data.
            - user (str): User ID for whom the action space will be plotted.
            - how (str): Plotting method ('vignette' or 'folium').
            - save (bool): If True, save the plot.

        Returns:
            - Visualization of the action space.
        """

        cols = [
            "user_id",
            "lon",
            "lat",
            "cluster_size",
            "cluster_info",
            "location_id",
            "home_location_id",
            "num_trip",
            "main_home_location_id",
            "dist_from_home",
            "imputed_purpose",
        ]

        # Determine aggregation method and check if it exists in the DataFrame
        if "user_id" in action_space.columns or "user_id" == action_space.index.name:
            aggregation_method = "user_id"
        elif (
            "user_id_day" in action_space.columns
            or "user_id_day" == action_space.index.name
        ):
            aggregation_method = "user_id_day"
        else:
            raise ValueError("No user_id_day or user_id column in act DataFrame")

        # check if user exists in the DataFrame
        if user not in act[aggregation_method].unique():
            raise ValueError(f"User {user} not in the DataFrame")

        act_ = act.loc[act[aggregation_method] == user, cols].copy()

        act_.drop(columns=["num_trip", "dist_from_home"], inplace=True)

        # Assuming your DataFrame is called df
        act_ = gpd.GeoDataFrame(
            act_, geometry=gpd.points_from_xy(act_["lon"], act_["lat"]), crs="EPSG:4326"
        )

        # Change the CRS to EPSG:2056
        act_ = act_.to_crs(
            "EPSG:2056"
        )  # This should be a parameter to change in the method, but default set to this one

        # Now, you can access the transformed x and y coordinates
        act_["x"], act_["y"] = act_.geometry.x, act_.geometry.y

        # And drop duplicated locations (information of visit count is in cluster_size)
        act_ = act_.drop_duplicates(ignore_index=True)

        if how == "vignette":
            self._plot_action_space_vignette(act_, action_space, user, save)
        elif how == "folium":
            return self._plot_action_space_folium(act_, action_space, user, save)
        else:
            raise ValueError(f"how must be either vignette or folium, not {how}")

    def _plot_action_space_vignette(
        self,
        act_: pd.DataFrame,
        action_space: pd.DataFrame,
        user: str,
        save: bool = False,
    ):
        """
        Plot the action space for a specific user using a vignette style.

        Args:
            - act_ (pandas.DataFrame): DataFrame with user activity data.
            - action_space (pandas.DataFrame): DataFrame with action space data.
            - user (str): User ID for whom the action space will be plotted.
            - save (bool): If True, save the plot.

        Returns:
            - Visualization of the action space using a vignette style.
        """

        # Create a GeoDataFrame for the points from act_
        geometry = [Point(x, y) for x, y in zip(act_["x"], act_["y"])]
        gdf = gpd.GeoDataFrame(act_, geometry=geometry, crs="EPSG:2056")

        # Create a Matplotlib figure
        fig, ax = plt.subplots(figsize=(10, 10))

        # Plot Ellipse

        e = action_space.loc[user, "ellipse_2056"]

        ellipse_polygon = Polygon(e.get_verts())
        ellipse_coords = np.array(ellipse_polygon.exterior.coords.xy)

        # ellipse_coords_4326 = np.column_stack(transformer.transform(x, y) for x, y in zip(ellipse_coords[0], ellipse_coords[1]))
        ax.fill(
            ellipse_coords[0],
            ellipse_coords[1],
            color="red",
            alpha=0.2,
            edgecolor="red",
        )

        # Plot points from act_
        ax.scatter(
            gdf["x"],
            gdf["y"],
            color="red",
            marker="o",
            s=50,
            label="Points from act",
        )

        # Plot wmc_ in green
        ax.scatter(
            action_space.loc[user, "wmc_x"],
            action_space.loc[user, "wmc_y"],
            color="green",
            marker="o",
            s=50,
            label="Weighted mean center",
        )
        find_home_geom = act_.loc[
            act_.location_id == act_.home_location_id.unique()[0], "geometry"
        ]
        if len(find_home_geom) == 1:
            home_loc = np.array([find_home_geom.x, find_home_geom.y], dtype=float)
        elif len(find_home_geom) > 1:
            home_loc = np.array([find_home_geom[0].x, find_home_geom[0].y], dtype=float)
        # Plot home_loc from act_ with a customized icon
        ax.scatter(
            home_loc[0],
            home_loc[1],
            color="black",
            marker="^",
            s=100,
            label="Home Location",
        )

        # Add legend
        ax.legend()

        # Set the extent of the plot based on the bounding box of the GeoDataFrame
        minx, miny, maxx, maxy = gdf.total_bounds
        ax.set_xlim(minx * 0.998, maxx * 1.002)
        ax.set_ylim(miny * 0.998, maxy * 1.002)

        # Plot basemap using OpenStreetMap
        ctx.add_basemap(ax, crs="EPSG:2056", source=ctx.providers.OpenStreetMap.Mapnik)

        # Save and show the figure
        if save:
            plt.savefig("action_space_vignette_" + user + ".png", dpi=300)

        plt.show()

    def _plot_action_space_folium(
        self,
        act_: pd.DataFrame,
        action_space: pd.DataFrame,
        user: str,
        save: bool = False,
    ):
        """
        Plot the action space for a specific user using Folium for interactive maps.

        Args:
            - act_ (pandas.DataFrame): DataFrame with user activity data.
            - action_space (pandas.DataFrame): DataFrame with action space data.
            - user (str): User ID for whom the action space will be plotted.
            - save (bool): If True, save the interactive map.

        Returns:
            - Folium interactive map displaying the action space for the given user.
        """

        # Find weighted mean center of a marked point pattern.
        wmc_ = weighted_mean_center(act_[["x", "y"]], act_.cluster_size)

        # Convert wmc_ to 'EPSG:4326'
        wmc_4326 = (
            gpd.GeoSeries(Point(wmc_[0], wmc_[1]), crs="EPSG:2056")
            .to_crs("EPSG:4326")
            .iloc[0]
        )
        wmc_4326 = [wmc_4326.y, wmc_4326.x]  # Flip lat and lon to match Folium format

        # Create a GeoDataFrame for the points from act_
        geometry = [Point(lon, lat) for lon, lat in zip(act_["lon"], act_["lat"])]
        gdf = gpd.GeoDataFrame(act_, geometry=geometry, crs="EPSG:4326")

        # Find the main home coordinates
        find_home_geom = act_.loc[
            act_.location_id == act_.home_location_id.unique()[0], "geometry"
        ]
        if len(find_home_geom) == 1:
            home_loc = np.array([find_home_geom.x, find_home_geom.y], dtype=float)
        elif len(find_home_geom) > 1:
            home_loc = np.array([find_home_geom[0].x, find_home_geom[0].y], dtype=float)

        # Create a GeoDataFrame for the home_loc
        home_loc_geometry = [Point(home_loc[0], home_loc[1])]
        home_loc_gdf = gpd.GeoDataFrame(
            geometry=home_loc_geometry, crs="EPSG:2056"
        ).to_crs("EPSG:4326")

        # Create a GeoDataFrame for the ellipse
        # Define the source and target coordinate reference systems (CRS)
        source_crs = "EPSG:2056"
        target_crs = "EPSG:4326"
        # Create a PyProj transformer for the conversion
        transformer = pyproj.Transformer.from_crs(
            source_crs, target_crs, always_xy=True
        )

        # Create Ellipse
        e = action_space.loc[user, "ellipse_2056"]
        ellipse_polygon = Polygon(e.get_verts())
        ellipse_coords = [[x, y] for x, y in ellipse_polygon.exterior.coords]
        # Convert coordinates to EPSG:4326
        ellipse_coords_4326 = [transformer.transform(x, y) for x, y in ellipse_coords]
        ellipse_coords_4326_t = [[y, x] for x, y in ellipse_coords_4326]
        ellipse_gdf = gpd.GeoDataFrame(
            geometry=[Polygon(ellipse_coords_4326)], crs="EPSG:4326"
        )

        # Create a Folium map centered around the mean coordinates
        mean_lat, mean_lon = gdf["lat"].mean(), gdf["lon"].mean()
        map_center = [mean_lat, mean_lon]
        mymap = folium.Map(location=map_center, zoom_start=12, control_scale=True)

        # Add basemap using OpenStreetMap
        folium.TileLayer("openstreetmap").add_to(mymap)

        # Add Ellipse
        # folium.Polygon(locations=ellipse.exterior.coords, color='#c9cfbc', fill=True, fill_color='#c9cfbc', fill_opacity=0.2).add_to(mymap)
        folium.GeoJson(
            ellipse_gdf,
            name="polygon",
            zoom_on_click=True,
            style_function=lambda x: {"color": "red"},
        ).add_to(mymap)

        # Add points from act_ with CircleMarkers
        for idx, row in gdf.iterrows():
            folium.CircleMarker(
                location=[row["lat"], row["lon"]],
                radius=5,
                color="red",
                fill=True,
                fill_color="red",
                fill_opacity=0.7,
                popup=f"{row['user_id']}",
            ).add_to(mymap)

        # Add wmc_ in green
        folium.CircleMarker(
            location=wmc_4326,
            radius=5,
            color="green",
            fill=True,
            fill_color="green",
            fill_opacity=1,
            popup="Weighted mean ceter",
        ).add_to(mymap)

        # Add home_loc from act_ with a customized icon
        folium.Marker(
            location=[home_loc_gdf.geometry.y.iloc[0], home_loc_gdf.geometry.x.iloc[0]],
            icon=folium.Icon(color="red", icon="home"),
        ).add_to(mymap)

        if save:
            mymap.save("action_space_folium_" + user + ".html")

        # Display the map
        return mymap

    # def get_inno_rate_per_phase(
    #     action_space,
    #     multiday_graph,
    #     user_id_,
    #     col_digraph_motif="DiGraph_motif",
    #     phase=None,
    #     treatment=None,
    # ):
    #     """
    #     Compute the innovation rate from motif
    #     user_id: list of users to compute the innovation rate
    #     mtf_useridday: df of DiGraph_motifs with user_id_days in index
    #     """

    #     mtf_useridday = pd.merge(
    #         multiday_graph[["DiGraph_motif", "motif_flat"]],
    #         action_space,
    #         left_index=True,
    #         right_index=True,
    #         how="left",
    #     )

    #     # df_innov_rate = pd.DataFrame(index=range(100),columns=user_id)

    #     # isolate rows for phase X only
    #     if phase != None:
    #         mtf_useridday = mtf_useridday[mtf_useridday.phase == phase]
    #     # isolate rows for treatment X only
    #     if treatment != None:
    #         mtf_useridday = mtf_useridday[mtf_useridday.treatment == treatment]
    #     # resample for continuous days
    #     graphs = (
    #         mtf_useridday[mtf_useridday.user_id == user_id_]
    #         .set_index("date")
    #         .resample("D")
    #         .ffill()
    #     )
    #     # read the first graph G0 for user_id
    #     G0 = graphs[col_digraph_motif][0]
    #     # init inno rate list
    #     innovation_rate = []
    #     # then compute the next graph and compare with G0
    #     for G in graphs[col_digraph_motif][1:]:
    #         innovation_rate.append(nx.compose(G0, G).number_of_nodes())
    #         G0 = nx.compose(G0, G)
    #     # x = np.arange(0, len(innovation_rate), 1).tolist()
    #     # y = innovation_rate

    #     return innovation_rate



__version__ = "0.1.0"
from .fake_data_generator import FakeDataGenerator
from .gps_analytics import GPSAnalytics
from .gps_data_privacy import GPSDataPrivacy
from .gps_data_processor import GPSDataProcessor
from .gps_to_actionspace import GPStoActionspace
from .gps_to_graph import GPStoGraph
from .xyt_plot import plot_gps_on_map



import pandas as pd
from faker import Faker
import random
from shapely.geometry import Point, LineString
import pytz
import geopandas as gpd
import pyproj
from geopy.geocoders import Nominatim
import folium
from shapely.wkt import loads
from shapely.geometry import Point
from IPython.display import display
from datetime import timedelta
from geopy.distance import geodesic
import numpy as np
from datetime import datetime, timedelta
import geopy.distance


class FakeDataGenerator:
    """
    A class for generating fake location-based data for testing and demonstration purposes.

    Attributes:
        - projection_epsg (int): The EPSG code for the desired projection (default is 4326).
        - location_name (str): The name of a location for which to generate data. If provided, the generated data will be constrained within the bounding box of this location.
        - home_radius_km (int): The radius in kilometers within which home locations for different groups should be generated.

    Example usage:
        .. code-block:: python

            generator = FakeDataGenerator(location_name="Switzerland", num_users=15)
            df_legs = generator.generate_legs(num_rows=10)
            df_staypoints = generator.generate_staypoints(num_rows=10)
            df_waypoints = generator.generate_waypoints(num_rows=10)


    """

    def __init__(
        self, projection_epsg=4326, location_name=None, home_radius_km=20, num_users=15
    ):
        self.fake = Faker()
        self.timezone = pytz.timezone("Europe/Zurich")
        self.projection_epsg = projection_epsg
        self.location_name = location_name
        self.projection = pyproj.Transformer.from_crs(
            "epsg:4326", f"epsg:{projection_epsg}", always_xy=True
        ).transform
        self.location_bbox = None
        self.home_radius_km = home_radius_km
        self.num_users = num_users

        if location_name:
            self.location_bbox = self.get_bounding_box(location_name)

    def get_bounding_box(self, location_name):
        """
        Retrieve the bounding box coordinates for a specified location.

        Args:
            - location_name (str): The name of the location for which to retrieve the bounding box.

        Returns:
            - list: A list of bounding box coordinates [south, north, west, east].
        """
        geolocator = Nominatim(user_agent="geo_data_generator")
        location = geolocator.geocode(location_name, exactly_one=True)

        if location is None:
            raise ValueError(f"Location '{location_name}' not found")

        return location.raw["boundingbox"]

    def generate_home_location(self):
        """
        Generate a random home location point.

        Returns:
            - Point: A Shapely Point object representing the home location.
        """
        if self.location_bbox:
            south, north, west, east = map(float, self.location_bbox)
            home_lon = random.uniform(float(west), float(east))
            home_lat = random.uniform(float(south), float(north))
            return Point(home_lon, home_lat)
        else:
            home_lon, home_lat = self.fake.longitude(), self.fake.latitude()
            return Point(home_lon, home_lat)

    def generate_random_point_in_location(self, home_location=None):
        """
        Generate a random point within the specified location.

        Args:
            - home_location (Point, optional): The home location within which to generate the random point.

        Returns:
            - Point: A Shapely Point object representing the random point within the location.
        """
        if home_location is None:
            home_location = self.generate_home_location()

        lat_offset, lon_offset = np.random.uniform(-1, 1, 2) * (
            self.home_radius_km / 111
        )

        random_lat = home_location.y + lat_offset
        random_lon = home_location.x + lon_offset

        return Point(random_lon, random_lat)

    def generate_intermediate_points(
        self, point_start, point_end, num_points, noise_range=(0.0001, 0.005)
    ):
        """
        Generate intermediate points between two given points.

        Args:
            - point_start (Point): The starting point.
            - point_end (Point): The ending point.
            - num_points (int): The number of intermediate points to generate.
            - noise_range (tuple, optional): Range for generating random noise (default is (0.0001, 0.005)).

        Returns:
            - list: A list of Shapely Point objects representing intermediate points.
        """
        x_start, y_start = point_start.x, point_start.y
        x_end, y_end = point_end.x, point_end.y

        intermediate_points = []
        for i in range(num_points):
            fraction = (i + 1) / (num_points + 1)
            lon = x_start + (x_end - x_start) * fraction
            lat = y_start + (y_end - y_start) * fraction

            # Generate random noise within the specified range
            noise = random.uniform(noise_range[0], noise_range[1])
            lon += random.uniform(-noise, noise)
            lat += random.uniform(-noise, noise)

            intermediate_points.append(Point(lon, lat))
        return intermediate_points

    def generate_dataframe(self, num_rows, trace_type):
        """
        Generate a GeoDataFrame with synthetic location-based data for multiple groups.

        Args:
            - num_rows (int): The number of data rows to generate for each group.
            - trace_type (str): The type of data to generate ('Staypoint' or 'Leg').

        Returns:
            - GeoDataFrame: A GeoDataFrame containing the generated location-based data.
        """
        data = []
        home_locations = []

        for group_id in range(self.num_users):
            id = self.fake.uuid4()
            home_location = self.generate_home_location()
            home_locations.extend([home_location] * num_rows)
            start_time = self.fake.date_time_this_decade()
            finished_at = start_time + timedelta(minutes=1)

            for _ in range(num_rows):
                if trace_type == "Staypoint":
                    start_time = finished_at + timedelta(minutes=random.randint(3, 60))
                    finished_at = start_time + timedelta(
                        minutes=random.randint(60, 600)
                    )
                elif trace_type == "Leg":
                    start_time = finished_at + timedelta(
                        minutes=random.randint(60, 600)
                    )
                    finished_at = start_time + timedelta(minutes=random.randint(3, 60))
                detected_mode = (
                    f"Mode::{random.choice(['Car', 'Transit', 'Walk', 'Bicycle'])}"
                )
                purpose = random.choice(["home", "work", "leisure", "other"])

                if trace_type == "Staypoint":
                    geometry = self.generate_random_point_in_location(home_location)
                elif trace_type == "Leg":
                    point_start = self.generate_random_point_in_location(home_location)
                    point_end = self.generate_random_point_in_location(home_location)

                    num_intermediate_points = random.randint(50, 200)
                    intermediate_points = self.generate_intermediate_points(
                        point_start, point_end, num_intermediate_points
                    )

                    geometry = LineString(
                        [point_start, *intermediate_points, point_end]
                    )

                    if point_start.x > 180 or point_start.y > 90:
                        estimated_length = geodesic(
                            (point_start.y, point_start.x), (point_end.y, point_end.x)
                        ).m
                    else:
                        estimated_length = point_start.distance(point_end)

                row = {
                    "user_id": id,
                    "type": trace_type,
                    "started_at": start_time,
                    "finished_at": finished_at,
                    "timezone": "Europe/Zurich",
                    "length_meters": None,
                    "detected_mode": detected_mode if trace_type == "Leg" else None,
                    "validated": bool(random.getrandbits(1)),
                    "purpose": purpose,
                    "geometry": geometry,
                    "started_on": start_time.date(),
                }
                data.append(row)

        df = pd.DataFrame(data)
        df["home_location"] = home_locations
        gdf = gpd.GeoDataFrame(df, crs=self.projection_epsg, geometry=df.geometry)
        if trace_type == "Leg":
            gdf["length_meters"] = round(gdf.geometry.to_crs("EPSG:2056").length, 1)
        else:
            del gdf["length_meters"]
        return gdf

    def generate_legs(self, num_rows=15):
        """
        Generate synthetic location-based data representing movement between locations (legs) for multiple users.

        Args:
            - num_rows (int, optional): The number of data rows to generate for each user (default is 15).

        Returns:
            - GeoDataFrame: A GeoDataFrame containing the generated legs data.
        """
        return self.generate_dataframe(num_rows, trace_type="Leg")

    def generate_staypoints(self, num_rows=15):
        """
        Generate synthetic location-based data representing user staypoints at various locations.

        Args:
            - num_rows (int, optional): The number of data rows to generate for each user (default is 15).

        Returns:
            - GeoDataFrame: A GeoDataFrame containing the generated staypoints data.
        """
        return self.generate_dataframe(num_rows, trace_type="Staypoint")

    def generate_waypoints(
        self, num_rows=15, num_extra_od_points=10, max_displacement_meters=20
    ):
        """
        Generate synthetic location-based data representing waypoints with timestamps and accuracy for multiple users.

        Args:
            - num_rows (int, optional): The number of data rows to generate for each user (default is 15).
            - num_extra_od_points (int, optional): The number of extra points to add at the beginning and end of each linestring (default is 2).
            - max_displacement_meters (float, optional): The maximum displacement in meters for the extra points (default is 10 meters).

        Returns:
            - GeoDataFrame: A GeoDataFrame containing the generated waypoints data.
        """
        df_tracks = self.generate_dataframe(num_rows, trace_type="Leg")
        waypoint_data = []

        for _, row in df_tracks.iterrows():
            line_string = row["geometry"]
            start_time = row["started_at"]
            end_time = row["finished_at"]
            point_count = len(list(line_string.coords))
            time_differences = np.linspace(0, 1, point_count)
            accuracies = np.random.poisson(100, point_count) / np.random.uniform(1, 100)

            for i in range(num_extra_od_points):
                lat, lon = line_string.coords[0]
                original_point = (lat, lon)
                displacement = geopy.distance.distance(
                    meters=random.uniform(0, max_displacement_meters)
                ).destination(original_point, random.uniform(0, 360))
                lat, lon = displacement[0], displacement[1]
                timestamp = start_time - timedelta(seconds=(num_extra_od_points - i))
                accuracy = accuracies[0] + accuracies[0] * (np.random.uniform(0, 0.50))
                waypoint_data.append(
                    [row["user_id"], "Waypoint", timestamp, lat, lon, accuracy]
                )

            for i, point in enumerate(line_string.coords):
                lat, lon = point
                timestamp = end_time + time_differences[i] * (end_time - start_time)
                accuracy = accuracies[i]
                waypoint_data.append(
                    [row["user_id"], "Waypoint", timestamp, lat, lon, accuracy]
                )

            for i in range(1, num_extra_od_points + 1):
                lat, lon = line_string.coords[-1]
                original_point = (lat, lon)
                displacement = geopy.distance.distance(
                    meters=random.uniform(0, max_displacement_meters)
                ).destination(original_point, random.uniform(0, 360))
                lat, lon = displacement[0], displacement[1]
                timestamp = end_time + timedelta(seconds=i)
                accuracy = accuracies[-1] + accuracies[-1] * (
                    np.random.uniform(0, 0.50)
                )
                waypoint_data.append(
                    [row["user_id"], "Waypoint", timestamp, lat, lon, accuracy]
                )

        waypoint_columns = [
            "user_id",
            "type",
            "tracked_at",
            "latitude",
            "longitude",
            "accuracy",
        ]
        df_waypoints = pd.DataFrame(waypoint_data, columns=waypoint_columns)

        return df_waypoints



import pandas as pd
import networkx as nx
import random
import numpy as np
from matplotlib import pyplot as plt
import gif
import geopandas as gpd
import multiprocessing as mp
import os
from pathos.multiprocessing import ProcessingPool as Pool


class GPStoGraph:
    def __init__(self) -> None:
        pd.set_option("display.max_columns", 999)

    @staticmethod
    def verify_columns(df: pd.DataFrame, expected_columns: list[str]) -> None:
        """
        Checks for missing columns in a DataFrame.

        Args:
            - df (pd.DataFrame): DataFrame to check columns.
            - expected_columns (list[str]): List of expected column names.

        Raises:
            - ValueError: If columns in `expected_columns` are missing in the DataFrame.
        """

        missing_columns = [col for col in expected_columns if col not in df.columns]
        if missing_columns:
            raise ValueError(
                f"Columns missing in DataFrame: {', '.join(missing_columns)}"
            )

    @staticmethod
    def _preprocess_nodes(nodes) -> pd.DataFrame:
        """
        Preprocesses GPS data by removing consecutive duplicate location entries.

        Args:
            - nodes (pd.DataFrame): DataFrame containing GPS data.

        Returns:
            - pd.DataFrame: Preprocessed GPS data.
        """

        # Create a copy of the original DataFrame
        nodes_ = nodes.reset_index(drop=True).copy()

        # Shift the location_id to create a 'location_id_next' column
        nodes_["location_id_next"] = nodes_.groupby("user_id_day")["location_id"].shift(
            -1
        )

        # Identify consecutive duplicate locations and remove them
        index_to_drop = []
        i = 0
        for i_0 in nodes_.loc[
            nodes_.location_id == nodes_.groupby("user_id_day")["location_id"].shift(-1)
        ].index:
            if i_0 <= i:
                index_to_drop.append(i_0)
                continue
            else:
                i = i_0
                location_id = nodes_.loc[i, "location_id"]

                while location_id == nodes_.loc[i, "location_id_next"]:
                    i += 1
                    index_to_drop.append(i)

                # Update 'finished_at' and 'location_id_next' for the first occurrence
                nodes_.loc[i_0, "finished_at"] = nodes_.loc[i, "finished_at"]
                nodes_.loc[i_0, "location_id_next"] = nodes_.loc[i, "location_id_next"]

        # Drop the identified duplicate locations and reset index
        nodes_.drop(index_to_drop, inplace=True)
        nodes_.reset_index(drop=True, inplace=True)

        # Create an 'edges' column with tuples of location_id and location_id_next
        nodes_["edges"] = list(zip(nodes_.location_id, nodes_.location_id_next))

        # Create a GeoDataFrame and parse geometry to a tuple of coordinates in the right projection
        nodes_ = gpd.GeoDataFrame(
            nodes_, geometry=gpd.points_from_xy(nodes_.lon, nodes_.lat), crs="EPSG:4326"
        )
        nodes_["coordinates"] = list(zip(nodes_.geometry.x, nodes_.geometry.y))

        return nodes_

    @staticmethod
    def _get_motifs(nodes) -> pd.DataFrame:
        """
        Computes motifs based on GPS data and constructs corresponding graphs.

        Args:
            - nodes (pd.DataFrame): DataFrame containing GPS data.

        Returns:
            - pd.DataFrame: DataFrame with computed motifs and corresponding graphs.
        """

        mtfs = pd.DataFrame(index=nodes.user_id_day.unique())
        mtfs["graph"] = np.nan
        mtfs["graph_flat"] = np.nan

        mtfs["graph"] = mtfs["graph"].astype("object")
        mtfs["graph_flat"] = mtfs["graph_flat"].astype("object")

        for user_id_ in mtfs.index:
            nodes_ = nodes.loc[nodes.user_id_day == user_id_]

            try:
                attributes = nodes_.drop_duplicates(
                    subset=["user_id_day", "location_id", "started_at", "finished_at"]
                )[
                    [
                        "user_id_day",
                        "location_id",
                        "coordinates",
                        "started_at",
                        "finished_at",
                    ]
                ]
                attributes.set_index("location_id", inplace=True)

                attributes_time = attributes.groupby(by=["location_id"]).aggregate(
                    {"started_at": list, "finished_at": list}
                )
                attributes_time = (
                    attributes_time.groupby(level=0)
                    .apply(
                        lambda attributes_time: attributes_time.xs(
                            attributes_time.name
                        ).to_dict()
                    )
                    .to_dict()
                )

                attributes_location = pd.DataFrame(
                    attributes["coordinates"].drop_duplicates()
                ).T.to_dict()

                G = nx.DiGraph()
                G.add_edges_from(nodes_.edges[:-1].to_list())

                nx.set_node_attributes(G, attributes_time, name="time")
                nx.set_node_attributes(G, attributes_location)

                mtfs.at[user_id_, "graph"] = G
                mtfs.at[user_id_, "graph_flat"] = (
                    nx.to_numpy_array(G).flatten().tolist()
                )

            except:
                print("Exception raised for " + str(user_id_))
                continue

        return mtfs.reset_index(drop=False)

    def get_graphs(self, df: pd.DataFrame, verbose: bool = True) -> pd.DataFrame:
        """
        Extracts motifs and graphs from GPS data.

        Args:
            - df (pd.DataFrame): DataFrame containing GPS data.
            - verbose (bool): Verbosity for multiprocessing (default is True).

        Returns:
            - pd.DataFrame: DataFrame with extracted motifs and graphs.
        """

        col_to_use = [
            "user_id",
            "user_id_day",
            "location_id",
            "started_at",
            "finished_at",
            "lon",
            "lat",
            "home_location_id",
        ]

        self.verify_columns(df, col_to_use)

        nodes = df[col_to_use].copy()

        nodes = nodes.sort_values(by=["user_id_day", "started_at"])
        nodes = nodes.drop_duplicates(
            subset=["user_id_day", "location_id", "started_at"]
        )
        nodes.reset_index(inplace=True, drop=True)

        mtfs = pd.DataFrame(index=nodes.user_id_day.unique())
        mtfs["mtf_loc"] = np.nan
        mtfs["mtf_loc"] = mtfs["mtf_loc"].astype("object")

        nodes_ = self._preprocess_nodes(nodes)

        cores = mp.cpu_count() - 1

        if verbose:
            print("Multiprocessing is launched with %s cores in parallel" % cores)
            print("...")

        # split the df in as many array as the machine has cores
        user_ids = np.array_split(nodes_.user_id_day.unique(), cores, axis=0)
        nodes_split = []
        for u in user_ids:
            nodes_split.append(nodes_.loc[nodes_.user_id_day.isin(u.tolist())])

        # create the multiprocessing pool
        pool = Pool(cores)

        # process the DataFrame by mapping function to each df across the pool
        df_out = np.vstack(pool.map(self._get_motifs, nodes_split))

        # return the df
        mtf = pd.DataFrame(
            df_out, columns=["user_id_day", "DiGraph_motif", "motif_flat"]
        )

        # close down the pool and join
        pool.close()
        pool.join()
        pool.clear()

        mtf.set_index("user_id_day", inplace=True)
        mtf["user_id"] = mtf.index.str[:5]

        if verbose:
            print("Job done !")

        return mtf

    @staticmethod
    @gif.frame
    def _plot_graph(mtf: pd.DataFrame, user_id_: str, i: int) -> plt.figure:
        """
        Generates a plot of the graph motif for a specific user.

        Args:
            - mtf (pd.DataFrame): DataFrame containing motif data.
            - user_id_ (str): User identifier.
            - i (int): Index of the user's data.

        Returns:
            - plt.figure: Figure with the plot of the graph motif.
        """

        G_all = nx.compose_all(
            mtf.loc[mtf.user_id == user_id_, "DiGraph_motif"].values.tolist()
        )

        user_id_day_ = mtf.loc[mtf.user_id == user_id_].index[i]

        G1 = mtf.loc[user_id_day_, "DiGraph_motif"]

        f = plt.figure(
            frameon=False, figsize=(10, 10), dpi=100
        )  # figsize=(10,10),dpi=50,

        # plot the backgroud graph
        pos = nx.get_node_attributes(G_all, "coordinates")
        nx.draw_networkx_edges(
            G_all,
            pos,
            edge_color="black",
            alpha=0.20,
            arrowstyle="->",
            arrowsize=15,
            width=1,
            connectionstyle="arc3,rad=+0.15",
        )  # wedge,shrink_factor=0.5
        nx.draw_networkx_nodes(
            G_all,
            pos,
            node_color="black",
            alpha=0.20,
            node_size=70,
            nodelist=list(G_all.nodes())[1:],
        )
        nx.draw_networkx_nodes(
            G_all,
            pos,
            node_color="#dbcc9c",
            alpha=0.90,
            node_size=100,
            nodelist=list(G_all.nodes())[:1],
        )

        # plot the new daily graph
        pos = nx.get_node_attributes(G1, "coordinates")
        nx.draw_networkx_edges(
            G1,
            pos,
            edge_color="black",
            alpha=0.90,
            arrowstyle="->",
            arrowsize=15,
            width=2,
            connectionstyle="arc3,rad=+0.15",
        )  # wedge,shrink_factor=0.5
        nx.draw_networkx_nodes(
            G1,
            pos,
            node_color="black",
            alpha=0.90,
            node_size=100,
            nodelist=list(G1.nodes())[1:],
        )

        return f

    def plot_graph(self, mtf: pd.DataFrame, path="") -> None:
        """
        Creates GIFs displaying the graph motif for multiple users.

        Args:
            - mtf (pd.DataFrame): DataFrame containing motif data.
            - path (str): Path to save the GIFs (default is '').
        """

        for user_id_ in random.choices(mtf.user_id.unique(), k=2):
            frames = []
            for i in range(0, len(mtf.loc[mtf.user_id == user_id_].index)):
                frames.append(self._plot_graph(mtf, user_id_, i))
            file_name = "%s.gif" % user_id_
            save_path = os.path.join(path, file_name)
            gif.save(frames, save_path, duration=500)

    def plot_motif(self, mtf: pd.DataFrame) -> None:
        """
        Plots motifs and their frequency distribution.

        Args:
            - mtf (pd.DataFrame): DataFrame containing motif data.
        """

        mtf_ = mtf.copy()
        motif_to_keep = 9
        mtf_["motif_id"] = 99

        for i_, mtf in enumerate(mtf_.motif_flat.value_counts().index[:motif_to_keep]):
            mtf_.loc[mtf_.motif_flat.apply(lambda x: x == mtf), "motif_id"] = i_ + 1

        mtf_list = mtf_.groupby("motif_id")["motif_flat"].agg(
            motif_flat=pd.Series.mode, count=pd.Series.count
        )
        mtf_list["count"] = mtf_list["count"] / mtf_list["count"].sum()

        f = plt.figure(
            frameon=False,
            figsize=(20, 2),
        )  # figsize=(10,10),dpi=50,

        for sublot, id_ in enumerate(mtf_list.index[:-1]):
            axis = f.add_subplot(
                1,
                9,
                sublot + 1,
                xticks=[],
                yticks=[],
                frame_on=False,
                title="MOTIF %s\n(%s%%)"
                % (id_, round(mtf_list.loc[id_, "count"] * 100, 1)),
            )  # title='motif %s (%d%%)'%(counter,expl_)
            motif = mtf_list.loc[id_, "motif_flat"]
            dim = np.sqrt(len(motif)).astype(int)
            motif_arr = np.asarray(motif).reshape((dim, dim))
            G = nx.DiGraph(motif_arr)
            pos = nx.circular_layout(G)
            nx.draw_networkx_edges(
                G,
                pos,
                edge_color="black",
                alpha=0.80,
                arrowstyle="->",
                arrowsize=15,
                width=1,
                connectionstyle="arc3,rad=+0.15",
            )  # wedge,shrink_factor=0.5
            nx.draw_networkx_nodes(
                G, pos, node_color="black", node_size=70, nodelist=list(range(1, dim))
            )
            nx.draw_networkx_nodes(
                G, pos, node_color="#dbcc9c", node_size=80, nodelist=[0]
            )

        plt.show()

    @staticmethod
    def _get_mtf_sequence(mtf_data, user_id, pad=31) -> np.array:
        """
        Retrieves a motif sequence for a specific user.

        Args:
            - mtf_data: DataFrame containing motif data.
            - user_id (str): User identifier.
            - pad (int): Length of the output sequence (default is 31).

        Returns:
            - np.array: Padded motif sequence for the specified user.
        """

        # Extract motif data for the specified user
        sequence = (
            mtf_data.loc[mtf_data["user_id"] == user_id, ["motif_id", "date"]]
            .set_index("date")
            .sort_index()
        )

        # Align the first observation on Monday
        pad_before = 0
        if len(sequence) > 0:
            pad_before = sequence.index.dayofweek[0]  # 0 is Monday, 1 is Tuesday, etc.

        # Resample the time series to have continuous days, fill in missing values with 0
        sequence = (
            sequence.resample("1D")
            .mean()
            .fillna(0)
            .astype(int)
            .reset_index(drop=True)
            .T
        )

        # Make all the series the same length by padding after
        pad_after_ = pad - sequence.shape[1] - pad_before
        pad_after = max(0, pad_after_)

        if sequence.shape[1] > pad:
            return np.pad(
                sequence.values.flatten(), pad_width=[pad_before, 0], mode="constant"
            )[:pad]
        else:
            return np.pad(
                sequence.values.flatten(),
                pad_width=[pad_before, pad_after],
                mode="constant",
            )[:pad]
        # Example usage:
        # motif_sequence = get_mtf_sequence(treatment_data, mtf_data, phase=1, user_id='your_user_id')

    def motif_sequence(self, mtf: pd.DataFrame, n_cols: int = 60) -> pd.DataFrame:
        """
        Generates motif sequences for users based on motif data.

        Args:
            - mtf (pd.DataFrame): DataFrame containing motif data.
            - n_cols (int): Number of columns for the motif sequence (default is 60).

        Returns:
            - pd.DataFrame: DataFrame with generated motif sequences for users.
        """

        mtf_ = mtf.copy()
        motif_to_keep = 9
        mtf_["motif_id"] = 99

        for i_, mtf in enumerate(mtf_.motif_flat.value_counts().index[:motif_to_keep]):
            mtf_.loc[mtf_.motif_flat.apply(lambda x: x == mtf), "motif_id"] = i_ + 1

        mtf_ = mtf_.reset_index()
        mtf_["date"] = mtf_["user_id_day"].str[-8:]
        mtf_["date"] = pd.to_datetime(mtf_["date"], format="%Y%m%d")

        mtf_seq = pd.DataFrame(index=mtf_.user_id.unique(), columns=range(n_cols))

        for usrs in mtf_.user_id.unique():
            try:
                mtf_seq.loc[usrs] = self._get_mtf_sequence(
                    mtf_, user_id=usrs, pad=n_cols
                ).tolist()
            except:
                print("Exception raised on user " + usrs + " / phase 1")
                continue

        return mtf_seq



import numpy as np
from haversine import haversine_vector, Unit
from datetime import datetime, timedelta
import pandas as pd


class GPSDataPrivacy:
    """
    A class for obfuscating and aggregating GPS data to enhance privacy.

    This class provides methods to obfuscate GPS data by adding noise, shifting coordinates,
    and aggregating data within defined cell centers. It also calculates utility metrics for
    evaluating the impact of obfuscation on the data.
    """

    def __init__(self) -> None:
        pd.options.mode.chained_assignment = None

    def _add_noise(self, point, radius=100, offset=30) -> (float, float):
        """
        Adds random uniform noise to the given point, in a radius = radius - offset.

        Args:
            - point (tuple): Points as a tuple (latitude, longitude) in degrees.
            - radius (int, optional): Radius of the point neighborhood. Defaults to 100.
            - offset (int, optional): Offset from the perimeter of the neighborhood circle,
            where the shifted point can't be located. Defaults to 30.

        Returns:
            - tuple: Shifted point expressed as (latitude, longitude)
        """

        adjusted_radius = radius - offset

        assert adjusted_radius > 0, "Offset must be bigger than radius."

        latitude_shift_km = np.random.uniform(-radius, radius) / 1000
        longitude_shift_km = np.random.uniform(-radius, radius) / 1000

        while (
            np.sqrt(latitude_shift_km**2 + longitude_shift_km**2)
            > (radius - offset) / 1000
        ):
            latitude_shift_km = np.random.uniform(-radius, radius) / 1000
            longitude_shift_km = np.random.uniform(-radius, radius) / 1000

        return self._shift_point(point, latitude_shift_km, longitude_shift_km)

    def _shift_point(
        self, point, latitude_shift_km, longitude_shift_km
    ) -> (float, float):
        """
        Shifts the coordinates of a given point.

        Args:
            - point (tuple): Starting point as (latitude, longitude), in degrees.
            - latitude_shift_km (float): Latitude offset in kilometers.
            - longitude_shift_km (float): Longitude offset in kilometers.

        Returns:
            - tuple: Shifted point expressed as (latitude, longitude)
        """

        lat = point[0]
        lon = point[1]

        lat_km = self._lat_to_km(lat) + latitude_shift_km
        lon_km = self._lon_to_km(lat, lon) + longitude_shift_km

        return (self._km_to_lat(lat_km), self._km_to_lon(lon_km, lat))

    @staticmethod
    def _lon_to_km(latitude, longitude) -> float:
        """
        Expresses given longitude in kilometers to the east.

        Args:
            - latitude (float): Latitude expressed in degrees.
            - longitude (float): Longitude expressed in degrees.

        Returns:
            - float: Longitude as kilometers to the east.
        """

        km_east = 111.320 * np.cos(np.deg2rad(latitude)) * longitude
        return km_east

    @staticmethod
    def _lat_to_km(latitude) -> float:
        """
        Expresses given latitude in kilometers to the north.

        Args:
            - latitude (float): Latitude in degrees.

        Returns:
            - float: Latitude expressed in kilometers to the north.
        """

        km_north = 110.574 * latitude
        return km_north

    @staticmethod
    def _km_to_lon(km_east, latitude) -> float:
        """
        Expresses given kilometers to the east in longitude degrees.

        Args:
            - km_east (float): Longitude expressed in km going east from longitude 0,
            at the given latitude.
            - latitude (float): Latitude in degrees.

        Returns:
            - float: Longitude in degrees.
        """

        lon = km_east / (111.320 * np.cos(np.deg2rad(latitude)))
        return lon

    @staticmethod
    def _km_to_lat(km_north) -> float:
        """
        Expresses latitude given in kilometers to the east in degrees.

        Args:
            - km_north (float): Latitude expressed in kilometers to the north.

        Returns:
            - float: Latitude in degrees.
        """

        return km_north / 110.574

    def _assign_cell_center(self, latitude, longitude, cell_size) -> (float, float):
        """
        Returns the closest cell center for the given coordinates,
        when the map is divided into a lattice with the supplied cell_size.

        Args:
            - latitude (float): Latitude as degrees.
            - longitude (float): Longitude as degrees.
            - cell_size (float): Cell size in kilometers.

        Returns:
            - tuple(float, float): Coordinates of the closest cell center on the map.
        """

        km_east = self._lon_to_km(latitude, longitude)
        km_north = self._lat_to_km(latitude)

        km_east = km_east - km_east % cell_size + cell_size / 2
        km_north = km_north - km_north % cell_size + cell_size / 2

        lat = self._km_to_lat(km_north)
        lon = self._km_to_lon(km_east, lat)

        return (lat, lon)

    def obfuscate(
        self, df, locations, radius=100, offset=30, mode="remove"
    ) -> (pd.DataFrame, list[tuple[float, float]]):
        """
        Obfuscates the regions of points given in 'locations' parameter
        by either removing all the points in their proximity or changing the location
        of these points to one noisy location in the proximity circle.

        Args:
            - df (pandas.DataFrame): DataFrame with 'latitude' and 'longitude' columns.
            - locations (list): List of locations given as (latitude, longitude) tuples.
            - radius (int, optional): Radius of the obfuscation circle. Defaults to 100.
            - offset (int, optional): Smallest distance from the perimeter of the obfuscation circle at which the location of interest must be located. Defaults to 30.
            - mode (str, optional): Obfuscation mode, can be either 'remove' or 'assign'. Defaults to 'remove'.

        Returns:
            - pandas.DataFrame: DataFrame with obfuscated regions.
        """

        if not isinstance(radius, (int, float)) or radius <= 0:
            raise ValueError("radius must be a positive integer or float.")

        if not isinstance(offset, int) or offset <= 0:
            raise ValueError("offset must be a positive integer.")

        if not isinstance(locations, list):
            raise ValueError(
                "locations must be a list of (latitude, longitude) tuples."
            )

        if mode not in ["remove", "assign"]:
            raise ValueError("mode must be either 'remove' or 'assign'")

        # shift the home location
        shifted_locations = []
        output_df = df.copy()

        for location in locations:
            if location is not None:
                shifted_location = self._add_noise(location, radius, offset)
                shifted_locations.append(shifted_location)

                output_df["location_dist"] = haversine_vector(
                    np.column_stack(
                        (output_df.latitude.values, output_df.longitude.values)
                    ),
                    np.tile(np.array(shifted_location), (output_df.shape[0], 1)),
                    Unit.METERS,
                )

                if mode == "remove":
                    output_df = output_df.loc[output_df["location_dist"] > radius]
                elif mode == "assign":
                    output_df.loc[
                        output_df.location_dist <= radius, "latitude"
                    ] = shifted_location[0]
                    output_df.loc[
                        output_df.location_dist <= radius, "longitude"
                    ] = shifted_location[1]
            else:
                shifted_locations.append(None)

        return (
            output_df.drop(columns=["location_dist"]),
            shifted_locations,
        )

    def get_obfuscation_utility(self, w_prepared, w_obfuscated, legs) -> float:
        """
        Calculates the ratio of legs affected by obfuscation to total legs.

        Args:
            - w_prepared (pandas.DataFrame): Smoothed and cleaned waypoints DataFrame.
            - w_obfuscated (pandas.DataFrame): Smoothed, cleaned, and obfuscated waypoints DataFrame.
            - legs (pandas.DataFrame): DataFrame that contains assembled legs.

        Returns:
            - [float]: Ratio of legs affected by obfuscation to total legs.
        """

        # Check if required columns are present in the input DataFrames
        required_columns_w_prepared = ["tracked_at", "latitude", "longitude"]
        if not all(
            column in w_prepared.columns for column in required_columns_w_prepared
        ):
            raise ValueError(
                "w_prepared DataFrame must have columns: 'tracked_at', 'latitude', 'longitude'."
            )

        required_columns_w_obfuscated = ["tracked_at", "latitude", "longitude"]
        if not all(
            column in w_obfuscated.columns for column in required_columns_w_obfuscated
        ):
            raise ValueError(
                "w_obfuscated DataFrame must have columns: 'tracked_at', 'latitude', 'longitude'."
            )

        required_columns_legs = ["started_at", "finished_at"]
        if not all(column in legs.columns for column in required_columns_legs):
            raise ValueError(
                "legs DataFrame must have columns: 'started_at', 'finished_at'."
            )

        if w_obfuscated.shape[0] != w_prepared.shape[0]:  # assume remove mode
            timestamps = w_prepared[
                ~w_prepared.tracked_at.isin(w_obfuscated.tracked_at)
            ].tracked_at
        else:  # assume assign mode
            timestamps = w_prepared[
                (w_prepared.latitude != w_obfuscated.latitude)
                | (w_prepared.longitude != w_obfuscated.longitude)
            ].tracked_at

        count = 0

        timestamps = timestamps.values

        starts = legs.started_at.values
        ends = legs.finished_at.values

        legs_size = legs.shape[0]

        for i in range(legs_size):
            points_affected = timestamps[
                (timestamps >= starts[i]) & (timestamps <= ends[i])
            ].shape[0]
            if points_affected > 0:
                count += 1

        return 1 - count / legs.shape[0]

    @staticmethod
    def _dt_floor(dt, delta) -> datetime:
        """
        Performs the floor operation on datetime values,
        with given timedelta as the base, e.g. 2021-01-01 12:21:47 with timedelta of 15s
        returns 2021-01-01 12:21:45.

        Args:
            - dt (datetime.datetime): Datetime value to be floored.
            - delta (datetime.timedelta): Timedelta used for the floor operation.

        Returns:
            - datetime.datetime: Floored datetime.
        """

        # dt needs to be python datetime
        dt = pd.Timestamp(dt).to_pydatetime()
        seconds = int((dt - datetime.min.replace(tzinfo=dt.tzinfo)).total_seconds())
        remainder = timedelta(
            seconds=seconds % delta.total_seconds(),
            microseconds=dt.microsecond,
        )
        return dt - remainder

    def aggregate(
        self, waypoints_df, cell_size=0.2, delta=timedelta(minutes=15)
    ) -> pd.DataFrame:
        """
        Aggregates users in timedeltas and cells on the map.
        Returns a DataFrame with the count of users in a given timedelta and cell.

        Args:
            - waypoints_df (pandas.DataFrame): DataFrame with 'latitude', 'longitude', 'user_id' and 'tracked_at' columns.
            - cell_size (float): Size of the square cells on the map, in kilometers.
            - delta (datetime.timedelta): Frequency for the time aggregation, e.g. 15 minutes.

        Returns:
            - pandas.DataFrame: DataFrame with 'tracked_at', 'cell_latitude', 'cell_longitude' and 'count' columns.
                The 'cell_latitude' and 'cell_longitude' columns give coordinates of the centers of cells on the map.
        """

        # Check if required columns are present in the input DataFrame
        required_columns = ["latitude", "longitude", "tracked_at", "user_id"]
        if not all(column in waypoints_df.columns for column in required_columns):
            raise ValueError(
                "Input DataFrame must have columns: 'latitude', 'longitude', 'tracked_at', 'user_id'."
            )

        # Check if cell_size is a positive float
        if not isinstance(cell_size, (float, int)) or cell_size <= 0:
            raise ValueError("cell_size must be a positive float or integer.")

        # Check if delta is a timedelta
        if not isinstance(delta, timedelta):
            raise ValueError("delta must be a timedelta.")

        df = waypoints_df[["latitude", "longitude", "tracked_at", "user_id"]]

        cell_centers = [
            self._assign_cell_center(point[0], point[1], cell_size)
            for point in np.column_stack((df.latitude.values, df.longitude.values))
        ]

        cell_latitudes = [x[0] for x in cell_centers]
        cell_longitudes = [x[1] for x in cell_centers]

        df["cell_latitude"] = cell_latitudes
        df["cell_longitude"] = cell_longitudes

        tracked_at = pd.to_datetime(df.tracked_at).values
        tzinfo = df.tracked_at.iloc[0].tzinfo

        tracked_at = [self._dt_floor(dt, delta) for dt in tracked_at]

        df["tracked_at"] = tracked_at
        df.tracked_at = df.tracked_at.dt.tz_localize(tzinfo)
        df = df[["user_id", "tracked_at", "cell_latitude", "cell_longitude"]]
        df = df.drop_duplicates()

        return (
            df.groupby(["tracked_at", "cell_latitude", "cell_longitude"])
            .agg("count")
            .rename(columns={"user_id": "count"})
        )



import folium
from folium import plugins
import geopandas as gpd
from shapely.geometry import Point, LineString
import random


def plot_gps_on_map(
    df, trace_type=None, home_col=None, work_col=None, geo_columns=None
):
    """
    Plot location-based data on a Folium map with different colors for each group.

    Args:
        - df (pd.DataFrame): A Pandas DataFrame containing the location-based data to be plotted.
        - trace_type (str, optional): The type of data to be plotted ('Stay', 'Track', or 'Waypoint'). If provided, only data of the specified trace type will be plotted.
        - home_col (str, optional): Name of the column containing the home coordinates. Default is None.
        - work_col (str, optional): Name of the column containing the work coordinates. Default is None.
        - geo_columns (str or list, optional): Name of the column(s) containing the latitude and longitude coordinates. Can be a string (e.g., 'geometry') or a list (e.g., ['latitude', 'longitude']). Default is None.

    Returns:
        - Interactive Folium map.
    """
    if geo_columns:
        if isinstance(geo_columns, str):
            geom_column = geo_columns
        elif isinstance(geo_columns, list) and len(geo_columns) == 2:
            df["geometry"] = df.apply(
                lambda row: Point(row[geo_columns[1]], row[geo_columns[0]]), axis=1
            )
            geom_column = "geometry"
        else:
            raise ValueError(
                "Invalid geo_columns parameter. It must be a string or a list of two strings."
            )
    elif "geometry" in df:
        geom_column = "geometry"
    elif "longitude" in df and "latitude" in df:
        df["geometry"] = df.apply(
            lambda row: Point(row["latitude"], row["longitude"]), axis=1
        )
        geom_column = "geometry"
    else:
        raise ValueError(
            "DataFrame must have 'geometry' column or 'longitude' and 'latitude' columns."
        )

    if df.empty:
        return None

    if isinstance(df[geom_column].iloc[0], Point):
        m = folium.Map(
            location=[df[geom_column].iloc[0].y, df[geom_column].iloc[0].x],
            zoom_start=10,
        )
    elif isinstance(df[geom_column].iloc[0], LineString):
        m = folium.Map(
            location=[
                df[geom_column].iloc[0].coords[0][1],
                df[geom_column].iloc[0].coords[0][0],
            ],
            zoom_start=10,
        )

    user_colors = {}  # To store colors for each user ID
    plotted_home = set()  # To track which users' home locations have been plotted
    plotted_work = set()  # To track which users' work locations have been plotted

    for index, row in df.iterrows():
        if trace_type and row["type"] != trace_type:
            continue  # Skip rows that don't match the specified trace type

        user_id = row["user_id"]
        if user_id not in user_colors:
            user_colors[user_id] = "#{:02x}{:02x}{:02x}".format(
                random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)
            )

        color = user_colors[user_id]
        popup_text = f"Type: {row['type']}<br>ID: {row['user_id']}"

        if row["type"] == "Staypoint":
            lat, lon = (
                row[geom_column].y,
                row[geom_column].x,
            )  # Swap latitude and longitude
            folium.CircleMarker(
                location=[lat, lon],
                color="red",
                radius=4,
                fill=True,
                fill_color="red",
                popup=popup_text,
            ).add_to(m)
        elif row["type"] == "Waypoint" or row["type"] == "Leg":
            if isinstance(row[geom_column], Point):
                lat, lon = (
                    row[geom_column].y,
                    row[geom_column].x,
                )  # Swap latitude and longitude
                folium.CircleMarker(
                    location=[lat, lon],
                    color=color,
                    radius=2,
                    fill=True,
                    fill_color=color,
                    popup=popup_text,
                ).add_to(m)
            elif isinstance(row[geom_column], LineString):
                linestring = [
                    (point[1], point[0]) for point in row[geom_column].coords
                ]  # Swap latitude and longitude
                folium.PolyLine(
                    locations=linestring, color=color, popup=popup_text
                ).add_to(m)

        if home_col and user_id not in plotted_home and home_col in df:
            home_location = df[df["user_id"] == user_id][home_col].iloc[0]
            folium.Marker(
                location=[home_location.y, home_location.x],
                icon=folium.Icon(color="red", icon="home", prefix="fa"),
            ).add_to(m)
            plotted_home.add(user_id)

        if work_col and user_id not in plotted_work and work_col in df:
            work_location = df[df["user_id"] == user_id][work_col].iloc[0]
            folium.Marker(
                location=[work_location.y, work_location.x],
                icon=folium.Icon(color="red", icon="briefcase", prefix="fa"),
            ).add_to(m)
            plotted_work.add(user_id)

    folium.plugins.Fullscreen().add_to(m)
    return m



import pandas as pd
import numpy as np
import pandas as pd
from sklearn.metrics import DistanceMetric
from sklearn.cluster import DBSCAN
import datetime
import multiprocessing as mp
from pathos.multiprocessing import ProcessingPool as Pool
from functools import partial
import math


class GPSAnalytics:
    """
    Performs analytics on GPS data, including splitting activities over midnight,
    spatial clustering, deriving metrics, and computing daily statistics.

    This class offers methods to manipulate GPS data stored in DataFrames. It provides
    functionalities to split activities that span midnight into two, perform spatial
    clustering using DBSCAN, derive various metrics related to trips and activities,
    and compute daily descriptive statistics.
    """

    def __init__(self) -> None:
        pd.set_option("display.max_columns", 999)
        self.expect_col_spt = [
            "activity_id",
            "started_at",
            "finished_at",
            "purpose",
            "user_id",
            "lon",
            "lat",
        ]
        self.expect_col_leg = [
            "leg_id",
            "started_at",
            "finished_at",
            "detected_mode",
            "mode",
            "user_id",
            "geometry",
            "next_activity_id",
            "length",
            "duration",
        ]
        self.expect_col_ext_stp = [
            # "leg_id",
            "started_at",
            "finished_at",
            # "detected_mode",
            # "mode",
            "user_id",
            # "geometry",
            # "next_activity_id",
            # "length",
            "duration",
            "cluster",
            "cluster_size",
            "cluster_info",
            "location_id",
            "peak",
            "first_dep",
            "last_arr",
            "home_loop",
            "daily_trip_dist",
            "num_trip",
            "max_dist",
            "min_dist",
            "max_dist_from_home",
            "dist_from_home",
            "home_location_id",
            "weekday",
        ]

    @staticmethod
    def verify_columns(df: pd.DataFrame, expected_columns: list[str]) -> None:
        """
        Verifies if expected columns are present in the DataFrame.

        Args:
            - df: DataFrame to check columns in.
            - expected_columns: List of expected column names.
        """

        missing_columns = [col for col in expected_columns if col not in df.columns]
        if missing_columns:
            raise ValueError(
                f"Columns missing in DataFrame: {', '.join(missing_columns)}"
            )

    def split_overnight(
        self, staypoint: pd.DataFrame, time_columns=["started_at", "finished_at"]
    ) -> pd.DataFrame:
        """
        Splits activities going over midnight into two activities.

        Description:
            1. Split activities that go over midnight into two activities
            2. Allocate the same geolocation and activity purpose to the splitted activity
            3. Compute the duration of the splitted activities

        Args:
            - staypoint: DataFrame containing activities to split. Columns: 'activity_id', 'started_at', 'finished_at', 'purpose', 'user_id', 'lon', 'lat'
            - time_columns: Columns for activity start and end times.

        Returns:
            - DataFrame with split activities and new 'duration' column.
        """

        self.verify_columns(df=staypoint, expected_columns=self.expect_col_spt)

        def split_activity(row):
            # Check if the activity spans midnight
            if row[time_columns[0]].date() != row[time_columns[1]].date():
                # Split the activity into two parts
                part1 = row.copy()
                part2 = row.copy()

                part1[time_columns[1]] = pd.to_datetime(
                    part1[time_columns[0]].date().strftime("%Y-%m-%d") + " 23:59:59"
                )
                part2[time_columns[0]] = pd.to_datetime(
                    part2[time_columns[1]].date().strftime("%Y-%m-%d") + " 00:00:01"
                )

                return pd.DataFrame([part1, part2])

            return pd.DataFrame([row])

        # Apply the split_activity function to each row and concatenate the result
        split_activities = pd.concat(
            staypoint.apply(split_activity, axis=1).tolist(), ignore_index=True
        )

        # Compute the duration in seconds
        split_activities["duration"] = (
            split_activities[time_columns[1]] - split_activities[time_columns[0]]
        ).dt.total_seconds() / 60

        return split_activities

    @staticmethod
    def _spatial_clustering(
        gdf,
        eps=300,
        minpts=2,
        lon_lat_columns=["lon", "lat"],
        user_id_col="user_id",
        purpose_col="imputed_purpose",
    ) -> pd.DataFrame:
        """
        Performs Density-based spatial clustering (DBSCAN) on GeoDataFrame.

        Desription:
            1. Use the Density-based spatial clustering of applications with noise (DBSCAN) method to aggregate neighboring nodes
            2. Label the clusters with -1 being noise, and 0, 1, ..., n the number of the cluster. NB. 0 is not necessarily the denser cluster
            3. Aggregate the lon/lat to the mean of all nodes in a same cluster
            4. NB. clustering done for a given user_id and a given imputed_purpose

        Args:
            - gdf: GeoDataFrame with Lon/Lat nodes. Columns: Lon/Lat columns defined by 'lon_lat_columns', 'user_id', 'imputed_purpose'
            - eps: Maximum distance between samples.
            - minpts: Number of nodes in a neighborhood.
            - lon_lat_columns: Columns for longitude and latitude.
            - user_id_col: Column name for user IDs.
            - purpose_col: Column name for activity purpose.

        Returns:
            - GeoDataFrame with clustered lon/lat and labels.
        """

        # parameterize DBSCAN
        eps_rad = eps / 3671000.0  # meters to radians
        db = DBSCAN(
            eps=eps_rad, min_samples=minpts, metric="haversine", algorithm="ball_tree"
        )
        # add a column for cluster labelling
        gdf["cluster"] = np.nan
        gdf["cluster_size"] = np.nan
        # initialize the output DF
        output = pd.DataFrame()

        # NB run DBSCAN per user_id and per activity purpose
        for user_id in gdf[user_id_col].unique():
            for purpose in gdf.loc[gdf[user_id_col] == user_id][purpose_col].unique():
                # compute DBSCAN using straight-line haversine distances
                sub_gdf = []
                sub_gdf = gdf.loc[
                    (gdf[user_id_col] == user_id) & (gdf[purpose_col] == purpose)
                ]
                sub_gdf = sub_gdf.copy(deep=True)
                sub_gdf.reset_index(inplace=True)

                # Perform DBSCAN clustering from features, and return cluster labels.
                cl = db.fit_predict(
                    np.deg2rad(sub_gdf[[lon_lat_columns[0], lon_lat_columns[1]]])
                )

                max_size = 0
                cl_size = 0
                for cluster in np.unique(cl):
                    sub_gdf.loc[(cl == cluster).tolist(), "cluster"] = cluster
                    if cluster != -1:
                        sub_gdf.loc[(cl == cluster), "cluster_size"] = len(
                            sub_gdf.loc[(cl == cluster)]
                        )
                    else:
                        sub_gdf.loc[(cl == cluster), "cluster_size"] = 1

                    if cluster != -1:
                        sub_gdf.loc[
                            sub_gdf.cluster == cluster, lon_lat_columns[0]
                        ] = sub_gdf[lon_lat_columns[0]][
                            sub_gdf.cluster == cluster
                        ].mean()
                        sub_gdf.loc[
                            sub_gdf.cluster == cluster, lon_lat_columns[1]
                        ] = sub_gdf[lon_lat_columns[1]][
                            sub_gdf.cluster == cluster
                        ].mean()

                output = pd.concat([output, sub_gdf], ignore_index=True)

        output.sort_values(by=["index"], inplace=True)
        output.set_index("index", inplace=True)
        return output

    @staticmethod
    def _cluster_info(
        df, threshold=0.5, user_id_col="user_id", purpose_col="imputed_purpose"
    ) -> pd.DataFrame:
        """
        Computes cluster labels and importance of visited places.

        Args:
            - df: DataFrame with cluster labels. Columns: 'user_id', 'imputed_purpose', 'cluster', 'cluster_size'
            - threshold: Threshold for frequent/occasional visits.
            - user_id_col: Column name for user IDs.
            - purpose_col: Column name for activity purpose.

        Returns:
            - DataFrame with labeled places.
        """

        output = pd.DataFrame()
        df["cluster_info"] = np.nan

        for user_id in df[user_id_col].unique():
            for purpose in df.loc[df[user_id_col] == user_id][purpose_col].unique():
                sub_gdf = df.loc[
                    (df[user_id_col] == user_id) & (df[purpose_col] == purpose)
                ]
                sub_gdf = sub_gdf.copy(deep=True)
                sub_gdf.reset_index(inplace=True)

                max_size = sub_gdf["cluster_size"][sub_gdf.cluster != -1].max()

                sub_gdf.loc[sub_gdf.cluster == -1, "cluster_info"] = "Visited once"
                sub_gdf.loc[
                    (sub_gdf.cluster_size == max_size) & (sub_gdf.cluster != -1),
                    "cluster_info",
                ] = "Most visited"
                sub_gdf.loc[
                    (sub_gdf.cluster_size >= max_size * threshold)
                    & (sub_gdf.cluster_info != "Most visited")
                    & (sub_gdf.cluster != -1),
                    "cluster_info",
                ] = "Frequent visit"
                sub_gdf.loc[
                    (sub_gdf.cluster_size < max_size * threshold)
                    & (sub_gdf.cluster_info != "Most visited")
                    & (sub_gdf.cluster != -1),
                    "cluster_info",
                ] = "Occasional visit"

                output = pd.concat([output, sub_gdf], ignore_index=True)
        output.sort_values(by=["index"], inplace=True)
        output.set_index("index", inplace=True)

        return output

    def spatial_clustering(self, staypoint: pd.DataFrame) -> pd.DataFrame:
        """
        Aggregates locations of most visited places.

        Args:
            - staypoint: DataFrame containing activity locations. Columns: 'activity_id', 'started_at', 'finished_at', 'purpose', 'user_id', 'lon', 'lat'

        Returns:
            - DataFrame with clustered locations and labels.
        """

        self.verify_columns(df=staypoint, expected_columns=self.expect_col_spt)

        new_spt = staypoint.copy()

        # Aggregate the locations of most visited places per user_id and per imputed_purpose
        new_spt = self._spatial_clustering(new_spt, purpose_col="purpose")

        # Label the Most Visited Places
        new_spt = self._cluster_info(new_spt, purpose_col="purpose")

        # Aggregate the 100m-neighbooring lon,lat pairs
        db = DBSCAN(
            eps=100 / 3671000, min_samples=2, metric="haversine", algorithm="ball_tree"
        )
        cl = db.fit_predict(np.deg2rad(new_spt[["lon", "lat"]]))
        for cluster in np.unique(cl):
            if cluster != -1:
                new_spt.loc[cl == cluster, "lon"] = new_spt.loc[
                    cl == cluster, "lon"
                ].mean()
                new_spt.loc[cl == cluster, "lat"] = new_spt.loc[
                    cl == cluster, "lat"
                ].mean()
        # Add location id for unique lon,lat pairs
        new_spt["location_id"] = np.nan
        for counter, lon, lat in (
            new_spt.groupby(["lon", "lat"], as_index=False)
            .size()[["lon", "lat"]]
            .itertuples(index=True)
        ):
            new_spt.loc[
                (new_spt.lon == lon) & (new_spt.lat == lat), "location_id"
            ] = counter
        return new_spt

    @staticmethod
    def _get_distance(origin, destination, od_matrix) -> float:
        """
        Calculate distance between two geographical coordinates.

        Args:
            - lon1: Longitude of the first point.
            - lat1: Latitude of the first point.
            - lon2: Longitude of the second point.
            - lat2: Latitude of the second point.

        Returns:
            - Distance between the points (in kilometers).
        """

        if math.isnan(origin) or math.isnan(destination):
            output = np.nan
        else:
            output = od_matrix.loc[origin, destination] * 1000
        return output

    def get_metrics(self, staypoint: pd.DataFrame, leg: pd.DataFrame) -> pd.DataFrame:
        """
        Computes additional variables and metrics.

        Args:
            - staypoint: DataFrame with activity data. Columns: 'activity_id', 'started_at', 'finished_at', 'purpose', 'user_id', 'lon', 'lat'
            - leg: DataFrame with leg data. Columns: 'leg_id', 'started_at', 'finished_at', 'detected_mode', 'mode', 'user_id', 'geometry', 'next_activity_id', 'length', 'duration'

        Returns:
            - DataFrame with computed metrics.
        """

        self.verify_columns(df=staypoint, expected_columns=self.expect_col_spt)
        self.verify_columns(df=leg, expected_columns=self.expect_col_leg)
        # Derive additional variables
        df = staypoint.copy()
        # DAILY USER_ID: Add user_ids per day
        df.insert(
            1,
            "user_id_day",
            df["user_id"]
            + "_"
            + df.started_at.dt.year.astype(str)
            + df.started_at.dt.month.astype(str).str.zfill(2)
            + df.started_at.dt.day.astype(str).str.zfill(2),
        )
        # PEAK HOURS: Add boolean if trip starts (i.e. activity ends) in peak hour
        morning_hours = range(6, 10)  # 6:00 AM to 9:59 AM
        noon_hours = range(12, 15)    # 12:00 PM to 2:59 PM
        evening_hours = range(16, 20) # 4:00 PM to 7:59 PM

        df["peak"] = "off_peak"  # Assuming initially all times are off-peak
        df.loc[
            df['finished_at'].dt.hour.isin(morning_hours),
            "peak"
        ] = "morning_peak"
        df.loc[
            df['finished_at'].dt.hour.isin(noon_hours),
            "peak"
        ] = "noon_peak"
        df.loc[
            df['finished_at'].dt.hour.isin(evening_hours),
            "peak"
        ] = "evening_peak"
        # Get the time of first departure / last arrival
        df["first_dep"] = np.nan
        df["last_arr"] = np.nan
        df["home_loop"] = 0
        df["daily_trip_dist"] = np.nan
        df["num_trip"] = np.nan
        df["max_dist"] = np.nan
        df["min_dist"] = np.nan
        df["max_dist_from_home"] = np.nan
        df["dist_from_home"] = np.nan
        df["home_location_id"] = np.nan

        location = (
            staypoint[["lon", "lat", "location_id"]]
            .copy()
            .sort_values("location_id")
            .reset_index(drop=True)
        )
        location.drop_duplicates(ignore_index=True, inplace=True)
        od_matrix_kms = pd.DataFrame(
            DistanceMetric.get_metric("haversine").pairwise(
                location[["lat", "lon"]].to_numpy()
            )
            * 6373,
            columns=location.location_id.unique(),
            index=location.location_id.unique(),
        )

        # RUN PARRALEL FUNCTIONS FUNC0. FUNC1 & FUNC2
        # BE CAREFUL THIS PART CAN BE LONG
        # MAKE A PROGRESS BAR ?

        df.rename(columns={"purpose": "imputed_purpose"}, inplace=True)

        for func in [self._func1, self._func2, self._func3]:
            cores = mp.cpu_count()
            # split the df in as many array as the machine has cores
            user_ids = np.array_split(df.user_id_day.unique(), cores, axis=0)
            df_split = []
            for u in user_ids:
                df_split.append(df.loc[df.user_id_day.isin(u.tolist())])
            # create the multiprocessing pool
            pool = Pool(cores)
            # process the DataFrame by mapping function to each df across the pool
            if func == self._func2:
                func2_partial = partial(self._func2, leg=leg)
                df_out = np.vstack(pool.map(func2_partial, df_split))
            elif func == self._func3:
                func3_partial = partial(self._func3, od_matrix_kms=od_matrix_kms)
                df_out = np.vstack(pool.map(func3_partial, df_split))
            else:
                df_out = np.vstack(pool.map(func, df_split))

            # return the df
            df = pd.DataFrame(df_out, columns=df.columns)

            # close down the pool and join
            pool.close()
            pool.join()
            pool.clear()

            if func == self._func2:
                # drop the days with only one obesrvation and small connection duration
                df.drop(
                    df[(df.first_dep.isna()) & (df.duration < 43200)].index,
                    inplace=True,
                )  # 43200sec is 12 hours
                # Add weekdays
                df["weekday"] = df.started_at.dt.weekday
                # SORT VALUES
                df.sort_values(
                    by=["user_id_day", "started_at"], inplace=True, ignore_index=True
                )
                df["cluster_size"] = df["cluster_size"].astype(int)

            if func == self._func3:
                df.reset_index(inplace=True, drop=True)
                for index, row in df.iterrows():
                    df.loc[index, "dist_from_home"] = self._get_distance(
                        row["location_id"], row["home_location_id"], od_matrix_kms
                    )
        return df

    @staticmethod
    def _func1(df) -> pd.DataFrame:
        import datetime
        import pandas as pd
        import numpy as np

        ## Clean up the first / last activity of the day
        # CASE 1. Manage cases where the first/last act is at home but started sometime in the morning/afternoon --> set started_at == 00:00:01 / finished_at == 23:59:59
        # Set a df with only the first/last activities of user-days (NB DO NOT RESET OR IGNORE INDEXES):
        first_act = df.loc[
            (df.drop_duplicates(subset=["user_id_day"], keep="first").index)
        ].copy()
        last_act = df.loc[
            (df.drop_duplicates(subset=["user_id_day"], keep="last").index)
        ].copy()

        case1f = first_act[
            (first_act.started_at.dt.time > datetime.time(0, 0, 1))
            & (first_act.started_at.dt.time < datetime.time(12, 0, 0))
            & (first_act.imputed_purpose == "Home")
        ]
        case1l = last_act[
            (last_act.finished_at.dt.time > datetime.time(12, 0, 0))
            & (last_act.finished_at.dt.time < datetime.time(23, 59, 59))
            & (last_act.imputed_purpose == "Home")
        ]
        df.loc[case1f.index, "started_at"] = pd.to_datetime(
            df.loc[case1f.index, "started_at"].dt.date.astype(str) + "T00:00:01Z",
            format="%Y-%m-%dT%H:%M:%SZ",
        )
        df.loc[case1l.index, "finished_at"] = pd.to_datetime(
            df.loc[case1l.index, "finished_at"].dt.date.astype(str) + "T23:59:59Z",
            format="%Y-%m-%dT%H:%M:%SZ",
        )
        # Recalculate the durations
        df.loc[case1l.index.append(case1f.index), "duration"] = (
            df.loc[case1l.index.append(case1f.index), "finished_at"]
            - df.loc[case1l.index.append(case1f.index), "started_at"]
        ) / np.timedelta64(1, "s")

        # CASE 2. Drop all the user_id_day for wich the first / last activity does not starts / ends at 00:00:01 / 23:59:59
        len_before = len(df)
        # Set a df with only the first/last activities of user-days (NB DO NOT RESET OR IGNORE INDEXES):
        first_act = df.loc[
            (df.drop_duplicates(subset=["user_id_day"], keep="first").index)
        ].copy()
        last_act = df.loc[
            (df.drop_duplicates(subset=["user_id_day"], keep="last").index)
        ].copy()

        # Clean the user_id_day with home only
        index_to_drop = []
        for user_id in df.user_id_day.unique():
            try:
                if (
                    len(df.loc[df.user_id_day == user_id, "location_id"].unique()) == 1
                ) & (
                    "Home"
                    in df.loc[df.user_id_day == user_id, "imputed_purpose"].unique()
                ):
                    index_to_drop.extend(
                        df.loc[(df.user_id_day == user_id)].index[1:].tolist()
                    )
                    df.loc[(df.user_id_day == user_id), "duration"] == df.loc[
                        (df.user_id_day == user_id), "duration"
                    ].sum()
                    df.loc[(df.user_id_day == user_id), "started_at"] == df.loc[
                        (df.user_id_day == user_id), "started_at"
                    ].min()
                    df.loc[(df.user_id_day == user_id), "finished_at"] == df.loc[
                        (df.user_id_day == user_id), "finished_at"
                    ].max()
            except ValueError:
                continue
        df = df.loc[~df.index.isin(index_to_drop)].copy()
        df.reset_index(inplace=True, drop=True)

        df = df.loc[
            (
                ~df.user_id_day.isin(
                    first_act.loc[
                        first_act.started_at.dt.time != datetime.time(0, 0, 1),
                        "user_id_day",
                    ].array
                )
            )
        ]
        df = df.loc[
            (
                ~df.user_id_day.isin(
                    last_act.loc[
                        last_act.finished_at.dt.time != datetime.time(23, 59, 59),
                        "user_id_day",
                    ].array
                )
            )
        ]
        # len_after = len(df)
        # print('Warning: clean up operation reduced the df lenght by -' + str("{:.1f}".format((len_before-len_after)*100/len_before)) + ' %')
        ##Most user-days start at home:
        # print('But note that there are still some weird cases like a day starting with Shopping. Those cases are however very few :' + "\n" + df.drop_duplicates(subset=['user_id_day'], keep='first').groupby(by='imputed_purpose').count()['user_id_day'].to_string())

        return df

    @staticmethod
    def _func2(df, leg) -> pd.DataFrame:
        import datetime
        import geopandas

        # Compute miscellaneous additional variables
        for user_id in df.user_id_day.unique():
            try:
                if len(df.loc[(df.user_id_day == user_id)]) > 1:
                    df.loc[(df.user_id_day == user_id), "first_dep"] = df.loc[
                        (df.user_id_day == user_id), "finished_at"
                    ].dt.time.min()
                    df.loc[(df.user_id_day == user_id), "last_arr"] = df.loc[
                        (df.user_id_day == user_id), "started_at"
                    ].dt.time.max()
                if (
                    sum(
                        df.loc[(df.user_id_day == user_id), "imputed_purpose"].isin(
                            ["Home", "home"]
                        )
                    )
                    > 1
                ):
                    df.loc[(df.user_id_day == user_id), "home_loop"] = (
                        sum(
                            df.loc[(df.user_id_day == user_id), "imputed_purpose"].isin(
                                ["Home", "home"]
                            )
                        )
                        - 1
                    )
                # find from legs the actual trip distance
                date = datetime.datetime.strptime(
                    user_id[-8:], "%Y%m%d"
                ).date()  # retrieve the date of the concerned activities from the user_id_day string
                condition1 = leg.next_activity_id.isin(
                    df.loc[(df.user_id_day == user_id), "activity_id"].tolist()
                )  # spot all the legs tracked to reach the activities
                condition2 = leg.started_at.dt.date == date  # match the dates
                df.loc[(df.user_id_day == user_id), "daily_trip_dist"] = leg.loc[
                    (condition1) & (condition2), "length"
                ].sum()
                # return the number of trips between activities
                df.loc[(df.user_id_day == user_id), "num_trip"] = (
                    len(df[df.user_id_day == user_id]) - 1
                )
            except ValueError:
                pass

        return df

    @staticmethod
    def _func3(df, od_matrix_kms) -> pd.DataFrame:
        from sklearn.metrics import DistanceMetric
        import numpy as np
        import math

        # Compute the distances between locations
        for user_id in df.user_id_day.unique():
            try:
                # Compute max/min distance between all activity locations
                od_pairs = df.loc[
                    df.user_id_day == user_id, ["lon", "lat"]
                ].drop_duplicates(ignore_index=True)
                od_dist = (
                    DistanceMetric.get_metric("haversine").pairwise(
                        od_pairs[["lat", "lon"]].to_numpy()
                    )
                    * 6373
                )
                if len(od_dist) > 1:
                    df.loc[df.user_id_day == user_id, "max_dist"] = (
                        od_dist.max().astype(int) * 1000
                    )
                    df.loc[df.user_id_day == user_id, "min_dist"] = (
                        od_dist[np.nonzero(od_dist)].min().astype(int) * 1000
                    )  # get the min among non-null values
                else:
                    df.loc[df.user_id_day == user_id, "max_dist"] = 0
                    df.loc[df.user_id_day == user_id, "min_dist"] = 0
                # Compute max distance from home
                home_locations = df.loc[
                    (df.user_id_day == user_id)
                    & (df.imputed_purpose.str.lower() == "home"),
                    ["location_id", "cluster_size"],
                ]
                if len(home_locations.location_id.unique()) > 1:
                    home_id = home_locations.loc[
                        home_locations.cluster_size.idxmax(), "location_id"
                    ]
                else:
                    home_id = home_locations.location_id.mean()
                if math.isnan(home_id) == False:
                    all_id = df.loc[
                        (df.user_id_day == user_id) & (df.location_id != home_id),
                        "location_id",
                    ]
                    df.loc[df.user_id_day == user_id, "max_dist_from_home"] = (
                        od_matrix_kms.loc[
                            od_matrix_kms.index.isin(all_id), home_id
                        ].max()
                        * 1000
                    )
                    df.loc[df.user_id_day == user_id, "home_location_id"] = home_id
            except ValueError:
                continue

        return df

    def get_daily_metrics(self, ext_staypoint: pd.DataFrame) -> pd.DataFrame:
        """
        Constructs a matrix of daily descriptive statistics.

        Args:
            - ext_staypoint: DataFrame with relevant columns. Columns: 'started_at', 'finished_at', 'user_id', 'duration', 'cluster', 'cluster_size', 'cluster_info', 'location_id', 'peak', 'first_dep', 'last_arr', 'home_loop', 'daily_trip_dist', 'num_trip', 'max_dist', 'min_dist', 'max_dist_from_home', 'dist_from_home', 'home_location_id', 'weekday'

        Returns:
            - DataFrame with computed daily profiles.
        """

        self.verify_columns(df=ext_staypoint, expected_columns=self.expect_col_ext_stp)
        # Select relevant columns
        daily_act = ext_staypoint[
            [
                "user_id_day",
                "first_dep",
                "last_arr",
                "home_loop",
                "daily_trip_dist",
                "peak",
                "num_trip",
                "max_dist_from_home",
                "weekday",
            ]
        ].copy()

        # Convert 'first_dep' and 'last_arr' to minutes since midnight
        daily_act.loc[daily_act.first_dep.notnull(), "first_dep"] = (
            (
                pd.to_datetime(
                    daily_act.loc[daily_act.first_dep.notnull(), "first_dep"],
                    format="%H:%M:%S",
                )
                - np.datetime64("1900-01-01")
            )
            .dt.total_seconds()
            .div(60)
            .astype(int)
        )
        daily_act.loc[daily_act.last_arr.notnull(), "last_arr"] = (
            (
                pd.to_datetime(
                    daily_act.loc[daily_act.last_arr.notnull(), "last_arr"],
                    format="%H:%M:%S",
                )
                - np.datetime64("1900-01-01")
            )
            .dt.total_seconds()
            .div(60)
            .astype(int)
        )

        # Remove duplicate rows based on 'user_id_day'
        daily_act.drop_duplicates(
            subset=["user_id_day"], keep="first", ignore_index=True, inplace=True
        )

        # Create new columns 'am_peak', 'pm_peak', and 'noon_peak'
        daily_act["am_peak"] = daily_act["pm_peak"] = daily_act["noon_peak"] = 0
        daily_act.loc[daily_act.peak == "morning_peak", "am_peak"] = 1
        daily_act.loc[daily_act.peak == "evening_peak", "pm_peak"] = 1
        daily_act.loc[daily_act.peak == "noon_peak", "noon_peak"] = 1

        # Drop the 'peak' column
        daily_act.drop("peak", inplace=True, axis=1)

        # Set 'user_id_day' as the index
        daily_act.set_index("user_id_day", inplace=True)

        return daily_act
